#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Interfaces for working with an ECP5 MSPI configuration flash. """

from amaranth import Signal, Module, Cat, Elaboratable, Instance, DomainRenamer, C


class ECP5ConfigurationFlashInterface(Elaboratable):
    """ Gateware that creates a connection to an MSPI configuration flash.

    Automatically uses appropriate platform resources; this abstracts away details
    necessary to e.g. drive the MCLK lines on an ECP5, which has special handling.

    I/O port:
        B: spi -- The platform 

        I: sck -- The serial clock to be delivered to the SPI flash.
        I: sdi -- The SDI line to be passed through to the target flash.
        O: sdo -- The SDO line read from the target flash.
        I: cs  -- Active-high chip select line.
    """

    def __init__(self, *, bus, use_cs=False):
        """ Params:
            bus    -- The SPI bus object to connect to.
            use_cs -- Whether or not the CS line should be passed through to the target device.
        """

        self.bus = bus
        self.use_cs = use_cs

        #
        # I/O port
        #
        self.sck = Signal()
        self.sdi = Signal()
        self.sdo = Signal()
        self.cs  = Signal()


    def elaborate(self, platform):
        m = Module()

        # Get the ECP5 block that's responsible for driving the MCLK pin,
        # and drive it using our SCK line.
        user_mclk = Instance('USRMCLK', i_USRMCLKI=self.sck, i_USRMCLKTS=0)
        m.submodules += user_mclk

        # Connect up each of our other signals.
        m.d.comb += [
            self.bus.sdi.o .eq(self.sdi),
            self.sdo       .eq(self.bus.sdo.i)
        ]
        
        if self.use_cs:
            m.d.comb += self.bus.cs.o.eq(self.cs)
        if hasattr(self.bus.cs, "oe"):
            m.d.comb += self.bus.cs.oe.eq(self.use_cs)
        
        return m


class FlashUIDReader(Elaboratable):
    """ Gateware that implements a simple SPI Flash unique ID reader, triggered on reset. 
    
    I/O ports:

        B: bus  -- The SPI bus used to communicate with the device.
        
        O: uid  -- Contains the retrieved ID after the read is finished.
        O: done -- A signal that gets asserted when the read operation finishes.
    """

    # Opcode to read the chip's unique ID.
    READ_UID = 0x4B
    
    def __init__(self, bus, clock_period=2, domain="sync"):
        assert clock_period & (clock_period - 1) == 0  # only powers of 2
        self._domain = domain
        self.period  = clock_period
        self.bus     = bus

        #
        # I/O port
        #
        self.uid     = Signal(64)
        self.done    = Signal()

    def elaborate(self, platform):
        m = Module()

        # Clock generation and clock edge strobes
        cycles   = Signal(range(self.period))
        sck_fall = Signal()
        sck_rise = Signal()
        m.d.sync += [
            sck_fall.eq(cycles == self.period - 1),     # falling edge
            sck_rise.eq(cycles == self.period//2 - 1),  # rising edge
        ]
        
        # Output shift register and bit counter
        shreg_o = Signal(8, init=self.READ_UID)
        count_o = Signal(range(128), init=8*(1+4+8)-1)  # bytes: 1 opcode, 4 padding, 8 id

        with m.FSM(domain=self._domain):

            with m.State("XFER"):
                m.d.sync += [
                    self.bus.sck .eq(cycles[-1]),
                    self.bus.sdi .eq(shreg_o[-1]),
                    self.bus.cs  .eq(1),
                ]
                m.d.sync += cycles.eq(cycles + 1)

                # Read logic: latch on rising edge
                with m.If(sck_rise):
                    m.d.sync += self.uid.eq(Cat(self.bus.sdo, self.uid[:-1]))

                # Write logic: setup on falling edge
                with m.If(sck_fall):
                    m.d.sync += [
                        shreg_o .eq(Cat(C(0,1), shreg_o[:-1])),
                        count_o .eq(count_o - 1),
                    ]
                    with m.If(count_o == 0):
                        m.d.sync += self.bus.cs.eq(0)
                        m.next = 'END'
            
            with m.State("END"):
                m.d.comb += self.done.eq(1)

        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m


#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test                                 import LunaUSBGatewareTestCase, usb_domain_test_case

from luna.gateware.usb.usb2.descriptor import GetDescriptorHandlerBlock, DeviceDescriptorCollection, StandardDescriptorNumbers

from usb_protocol.emitters.descriptors.standard import get_string_descriptor

class GetDescriptorHandlerBlockTest(LunaUSBGatewareTestCase):
    descriptors = DeviceDescriptorCollection()

    with descriptors.DeviceDescriptor() as d:
        d.bcdUSB             = 2.00
        d.idVendor           = 0x1234
        d.idProduct          = 0x4567
        d.iManufacturer      = "Manufacturer"
        d.iProduct           = "Product"
        d.iSerialNumber      = "ThisSerialNumberIsResultsInADescriptorLongerThan64Bytes"
        d.bNumConfigurations = 1

        with descriptors.ConfigurationDescriptor() as c:
            c.bmAttributes = 0xC0
            c.bMaxPower = 50

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber   = 0
                i.bInterfaceClass    = 0x02
                i.bInterfaceSubclass = 0x02
                i.bInterfaceProtocol = 0x01

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x81
                    e.bmAttributes     = 0x03
                    e.wMaxPacketSize   = 64
                    e.bInterval        = 11

    # Add a non-consecutive string descriptor.
    descriptors.add_descriptor(get_string_descriptor("nonconsecutive"), index=0xfe)

    # HID Descriptor (Example E.8 of HID specification)
    descriptors.add_descriptor(b'\x09\x21\x01\x01\x00\x01\x22\x00\x32')

    FRAGMENT_UNDER_TEST = GetDescriptorHandlerBlock
    FRAGMENT_ARGUMENTS = {"descriptor_collection": descriptors}

    def traces_of_interest(self):
        dut = self.dut
        return (dut.value, dut.length, dut.start_position, dut.start, dut.stall,
                dut.tx.ready, dut.tx.first, dut.tx.last, dut.tx.payload, dut.tx.valid)

    def _test_descriptor(self, type_number, index, raw_descriptor, start_position, max_length, delay_ready=0):
        """ Triggers a read and checks if correct data is transmitted. """

        # Set a defined start before starting
        yield self.dut.tx.ready.eq(0)
        yield

        # Set up request
        yield self.dut.value.word_select(1, 8).eq(type_number)  # Type
        yield self.dut.value.word_select(0, 8).eq(index)  # Index
        yield self.dut.length.eq(max_length)
        yield self.dut.start_position.eq(start_position)
        yield self.dut.tx.ready.eq(1 if delay_ready == 0 else 0)
        yield self.dut.start.eq(1)
        yield

        yield self.dut.start.eq(0)

        yield from self.wait_until(self.dut.tx.valid, timeout=100)

        if delay_ready > 0:
            for _ in range(delay_ready-1):
                yield
            yield self.dut.tx.ready.eq(1)
            yield

        max_packet_length = 64
        expected_data = raw_descriptor[start_position:]
        expected_bytes = min(len(expected_data), max_length-start_position, max_packet_length)

        if expected_bytes == 0:
            self.assertEqual((yield self.dut.tx.first), 0)
            self.assertEqual((yield self.dut.tx.last),  1)
            self.assertEqual((yield self.dut.tx.valid), 1)
            self.assertEqual((yield self.dut.stall),    0)
            yield

        else:
            for i in range(expected_bytes):
                self.assertEqual((yield self.dut.tx.first),   1 if (i == 0) else 0)
                self.assertEqual((yield self.dut.tx.last),    1 if (i == expected_bytes - 1) else 0)
                self.assertEqual((yield self.dut.tx.valid),   1)
                self.assertEqual((yield self.dut.tx.payload), expected_data[i])
                self.assertEqual((yield self.dut.stall),      0)
                yield

        self.assertEqual((yield self.dut.tx.valid), 0)

    def _test_stall(self, type_number, index, start_position, max_length):
        """ Triggers a read and checks if correctly stalled. """

        yield self.dut.value.word_select(1, 8).eq(type_number)  # Type
        yield self.dut.value.word_select(0, 8).eq(index)  # Index
        yield self.dut.length.eq(max_length)
        yield self.dut.start_position.eq(start_position)
        yield self.dut.tx.ready.eq(1)
        yield self.dut.start.eq(1)
        yield

        yield self.dut.start.eq(0)

        cycles_passed = 0
        timeout = 100

        while not (yield self.dut.stall):
            self.assertEqual((yield self.dut.tx.valid), 0)
            yield

            cycles_passed += 1
            if timeout and cycles_passed > timeout:
                raise RuntimeError(f"Timeout waiting for stall!")

    @usb_domain_test_case
    def test_all_descriptors(self):
        for type_number, index, raw_descriptor in self.descriptors:
            yield from self._test_descriptor(type_number, index, raw_descriptor, 0, len(raw_descriptor))
            yield from self._test_descriptor(type_number, index, raw_descriptor, 0, len(raw_descriptor), delay_ready=10)

    @usb_domain_test_case
    def test_all_descriptors_with_offset(self):
        for type_number, index, raw_descriptor in self.descriptors:
            if len(raw_descriptor) > 1:
                yield from self._test_descriptor(type_number, index, raw_descriptor, 1, len(raw_descriptor))

    @usb_domain_test_case
    def test_all_descriptors_with_length(self):
        for type_number, index, raw_descriptor in self.descriptors:
            if len(raw_descriptor) > 1:
                yield from self._test_descriptor(type_number, index, raw_descriptor, 0, min(8, len(raw_descriptor)-1))
                yield from self._test_descriptor(type_number, index, raw_descriptor, 0, min(8, len(raw_descriptor)-1), delay_ready=10)

    @usb_domain_test_case
    def test_all_descriptors_with_offset_and_length(self):
        for type_number, index, raw_descriptor in self.descriptors:
            if len(raw_descriptor) > 1:
                yield from self._test_descriptor(type_number, index, raw_descriptor, 1, min(8, len(raw_descriptor)-1))

    @usb_domain_test_case
    def test_all_descriptors_with_zero_length(self):
        for type_number, index, raw_descriptor in self.descriptors:
            yield from self._test_descriptor(type_number, index, raw_descriptor, 0, 0)

    @usb_domain_test_case
    def test_unavailable_descriptor(self):
        yield from self._test_stall(StandardDescriptorNumbers.STRING, 100, 0, 64)

    @usb_domain_test_case
    def test_unavailable_index_type(self):
        # Unavailable index in between
        yield from self._test_stall(0x10, 0, 0, 64)

        # Index after last used type
        yield from self._test_stall(0x42, 0, 0, 64)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Simple gateware debug console for LUNA. """


from amaranth        import Signal, Module, Cat, Elaboratable, Array


class DebugConsole(Elaboratable):
    """ Simple debug console gateware for LUNA.

        I: line_in[8][max_line_legth] -- The line to be rendered to the output stream.
        I: line_length -- The data length to be rendered t

    """

    def __init__(self, *, max_line_length=128):

        self.line_in     = Array(Signal(8) for _ in range(max_line_length))
        self.line_length = Signal(range(0, max_line_length + 1))


    def elaborate(self, platform):
        m = Module()

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Contains the gatware module necessary to interpret and generate low-level USB packets. """


import operator
import functools

from amaranth          import Signal, Module, Elaboratable, Cat, Array, Const
from amaranth.hdl.rec  import Record, DIR_FANIN, DIR_FANOUT

from .                 import USBSpeed, USBPacketID
from ..stream          import USBInStreamInterface, USBOutStreamInterface
from ...interface.utmi import UTMITransmitInterface

#
# Interfaces.
#


class HandshakeExchangeInterface(Record):
    """ Record that carries handshakes detected -or- generated between modules.

    Attributes
    ----------
    ack: Signal()
        When connected to a generator, pulsing this strobe will trigger generating of an ACK.
        When connected to a detector, this strobe will be pulsed when an ACK is detected from the host.
    nak: Signal()
        When connected to a generator, pulsing this strobe will trigger generating of an NAK.
        When connected to a detector, this strobe will be pulsed when an NAK is detected from the host.
    stall: Signal()
        When connected to a generator, pulsing this strobe will trigger generation of a STALL.
        Unused in a detector, currently.
    nyet: Signal()
        When connected to a generator, pulsing this strobe will trigger generation of a NYET.
        Unused in a detector, currently.

    Parameters
    ----------
    is_detector: bool
        If true, this will be considered an interface to a detector that identifies handshakes.
        Otherwise, this will be considered an interface to a generator that accepts handshake requests.
    """

    def __init__(self, *, is_detector):
        direction = DIR_FANOUT if is_detector else DIR_FANOUT

        super().__init__([
            ('ack',   1, direction),
            ('nak',   1, direction),
            ('stall', 1, direction),
            ('nyet',  1, direction),
        ])



class DataCRCInterface(Record):
    """ Record providing an interface to a USB CRC-16 generator.

    Attributes
    ----------
    start: Signal(), input to CRC generator
        Strobe that indicates that a new CRC computation should be started.
    crc: Signal(), output from CRC generator
        The current CRC-16 value; updated with each sent or received byte.
    """

    def __init__(self):
        super().__init__([
            ('start', 1,  DIR_FANIN),
            ('crc',   16, DIR_FANOUT)
        ])


class TokenDetectorInterface(Record):
    """ Record providing an interface to a USB token detector.

    Attributes
    ----------
    pid: Signal(4), detector output
        The Packet ID of the most recent token.
    address: Signal(7), detector output
        The address associated with the relevant token.
    endpoint: Signal(4), detector output
        The endpoint indicated by the most recent token.

    new_token: Signal(), detector output
        Strobe asserted for a single cycle when a new token packet has been received.
    ready_for_response: Signal(), detector output
        Strobe asserted for a single cycle one inter-packet delay after a token packet is complete.
        Indicates when the token packet can be responded to.

    frame: Signal(11), detector output
        The current USB frame number.
    new_frame: Signal(), detector output
        Strobe asserted for a single cycle when a new SOF has been received.

    is_in: Signal(), detector output
        High iff the current token is an IN.
    is_out: Signal(), detector output
        High iff the current token is an OUT.
    is_setup: Signal(), detector output
        High iff the current token is a SETUP.
    is_ping: Signal(), detector output
        High iff the current token is a PING.
    """

    def __init__(self):
        super().__init__([
            ('pid',                4, DIR_FANOUT),
            ('address',            7, DIR_FANOUT),
            ('endpoint',           4, DIR_FANOUT),
            ('new_token',          1, DIR_FANOUT),
            ('ready_for_response', 1, DIR_FANOUT),

            ('frame',             11, DIR_FANOUT),
            ('new_frame',          1, DIR_FANOUT),

            ('is_in',              1, DIR_FANOUT),
            ('is_out',             1, DIR_FANOUT),
            ('is_setup',           1, DIR_FANOUT),
            ('is_ping',            1, DIR_FANOUT),
        ])


class InterpacketTimerInterface(Record):
    """ Record providing an interface to our interpacket timer.

    See [USB2.0: 7.1.18] and the USBInterpacketTimer gateware for more information.

    Attributes
    ----------
    start: Signal(), input to timer
        Strobe that indicates when the timer should be started. Usually started at the end of an Rx or Tx event.

    tx_allowed: Signal(), output from timer
        Strobe that goes high when it's safe to transmit after an Rx event.
    tx_timeout: Signal(), output from timer
        Strobe that goes high when the transmit-after-receive window has passed.
    rx_timeout: Signal(), output from timer
        Strobe that goes high when the receive-after-transmit window has passed.
    """

    def __init__(self):
        super().__init__([
            ('start',      1, DIR_FANIN),

            ('tx_allowed', 1, DIR_FANOUT),
            ('tx_timeout', 1, DIR_FANOUT),
            ('rx_timeout', 1, DIR_FANOUT),
        ])


    def attach(self, *subordinates):
        """ Attaches subordinate interfaces to the given timer interface.

        Parameters
        ----------
        subordinates: [InterpacketTimerInterface, Signal]
            Each :class:`InterpacketTimerInterface` is provided will be fully connected to a given
            timer interface. Each ``Signal`` provided will be interpreted as a timer reset, and added
            to the list of all resets.
        """

        start_conditions = []
        fragments = []

        for subordinate in subordinates:

            # If this is an interface, add its start to our list of start conditions,
            # and propagate our timer outputs to it.
            if isinstance(subordinate, self.__class__):
                start_conditions.append(subordinate.start)
                fragments.extend([
                    subordinate.tx_allowed.eq(self.tx_allowed),
                    subordinate.tx_timeout.eq(self.tx_timeout),
                    subordinate.rx_timeout.eq(self.rx_timeout)
                ])

            # If it's a signal, connect it directly as a start signal.
            else:
                start_conditions.append(subordinate)

        # Merge all of our start conditions into a single start condition, and
        # then add that to our fragment list.
        start_condition = functools.reduce(operator.__or__, start_conditions)
        fragments.append(self.start.eq(start_condition))

        return fragments


#
# Gateware.
#

class USBTokenDetector(Elaboratable):
    """ Gateware that parses token packets and generates relevant events.

    Attributes
    ----------
    interface: TokenDetectorInterface
        The interface that contains token detection events, and information about detected tokens.
    speed: Signal(2), input
        Carries a ``USBSpeed`` constant identifying the device's current operating speed.
    address: Signal(7), input -or- output
        If :parameter:``filter_by_address`` is true, this is an input that filters our event detector so
        it only reports tokens directed at a given address.
        If ``filter_by_address`` is false, this is an output that contains the address of the most
        recent token.


    Parameters
    ----------
        utmi: UTMIInterface
            The UTMI bus to observe.
        filter_by_address: bool
            If true, this detector will only report events for the address supplied in the address[] field.
    """

    SOF_PID      = 0b0101
    TOKEN_SUFFIX =   0b01

    def __init__(self, *, utmi, filter_by_address=True, domain_clock=60e6, fs_only=False):
        self.utmi = utmi
        self.filter_by_address = filter_by_address
        self._domain_clock = domain_clock
        self._fs_only = fs_only

        #
        # I/O port
        #
        self.interface = TokenDetectorInterface()
        self.speed     = Signal(2)
        self.address   = Signal(7)


    @staticmethod
    def _generate_crc_for_token(token):
        """ Generates a 5-bit signal equivalent to the CRC check for the provided token packet. """

        def xor_bits(*indices):
            bits = (token[len(token) - 1 - i] for i in indices)
            return functools.reduce(operator.__xor__, bits)

        # Implements the CRC polynomial from the USB specification.
        return Cat(
             xor_bits(10, 9, 8, 5, 4, 2),
            ~xor_bits(10, 9, 8, 7, 4, 3, 1),
             xor_bits(10, 9, 8, 7, 6, 3, 2, 0),
             xor_bits(10, 7, 6, 4, 1),
             xor_bits(10, 9, 6, 5, 3, 0)
        )


    def elaborate(self, platform):
        m = Module()

        token_data       = Signal(11)
        current_pid      = Signal.like(self.interface.pid)

        # Instantiate a dedicated inter-packet delay timer, which
        # we'll use to generate our `ready_for_response` signal.
        #
        # Giving this unit a timer separate from a device's main
        # timer simplifies the architecture significantly; and
        # removes a primary source of timer contention.
        m.submodules.timer = USBInterpacketTimer(domain_clock=self._domain_clock, fs_only=self._fs_only)
        timer              = InterpacketTimerInterface()
        m.d.comb += m.submodules.timer.speed.eq(self.speed)

        # Generate our 'ready_for_response' signal whenever our
        # timer reaches a delay that indicates it's safe to respond to a token.
        m.submodules.timer.add_interface(timer)
        m.d.comb += self.interface.ready_for_response.eq(timer.tx_allowed)

        # Generate our convenience status signals.
        m.d.comb += [
            self.interface.is_in     .eq(self.interface.pid == USBPacketID.IN),
            self.interface.is_out    .eq(self.interface.pid == USBPacketID.OUT),
            self.interface.is_setup  .eq(self.interface.pid == USBPacketID.SETUP),
            self.interface.is_ping   .eq(self.interface.pid == USBPacketID.PING)
        ]

        # Keep our strobes un-asserted unless otherwise specified.
        m.d.usb += [
            self.interface.new_frame  .eq(0),
            self.interface.new_token  .eq(0)
        ]

        with m.FSM(domain="usb"):

            # IDLE -- waiting for a packet to be presented
            with m.State("IDLE"):
                with m.If(self.utmi.rx_active):
                    m.next = "READ_PID"


            # READ_PID -- read the packet's ID, and determine if it's a token.
            with m.State("READ_PID"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                with m.Elif(self.utmi.rx_valid):
                    is_normal_token = (self.utmi.rx_data[0:2] == self.TOKEN_SUFFIX)
                    is_ping_token   = (self.utmi.rx_data[0:4] == USBPacketID.PING)
                    is_valid_pid    = (self.utmi.rx_data[0:4] == ~self.utmi.rx_data[4:8])

                    # If we have a valid token, move to capture it.
                    # Note that we have two categories of token we'll accept: normal tokens (IN, OUT, SETUP, SOF),
                    # and our SPECIAL category tokens (e.g. PING), which have a separate PID suffix.
                    with m.If((is_normal_token | is_ping_token) & is_valid_pid):
                        m.d.usb += current_pid.eq(self.utmi.rx_data)
                        m.next = "READ_TOKEN_0"

                    # Otherwise, ignore this packet as a non-token.
                    with m.Else():
                        m.next = "IRRELEVANT"


            with m.State("READ_TOKEN_0"):

                # If our transaction stops, discard the current read state.
                # We'll ignore token fragments, since it's impossible to tell
                # if they were e.g. for us.
                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                # If we have a new byte, grab it, and move on to the next.
                with m.Elif(self.utmi.rx_valid):
                    m.d.usb += token_data.eq(self.utmi.rx_data)
                    m.next = "READ_TOKEN_1"


            with m.State("READ_TOKEN_1"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                # Once we've just gotten the second core byte of our token,
                # we can validate our checksum and handle it.
                with m.Elif(self.utmi.rx_valid):
                    expected_crc = self._generate_crc_for_token(
                        Cat(token_data[0:8], self.utmi.rx_data[0:3]))

                    # If the token has a valid CRC, capture it...
                    with m.If(self.utmi.rx_data[3:8] == expected_crc):
                        m.d.usb += token_data[8:].eq(self.utmi.rx_data)
                        m.next = "TOKEN_COMPLETE"

                    # ... otherwise, we'll ignore the whole token, as we can't tell
                    # if this token was meant for us.
                    with m.Else():
                        m.next = "IRRELEVANT"

            # TOKEN_COMPLETE: we've received a full token; and now need to wait
            # for the packet to be complete.
            with m.State("TOKEN_COMPLETE"):

                # Once our receive is complete, use the completed token,
                # and strobe our "new token" signal.
                with m.If(~self.utmi.rx_active):
                    m.next="IDLE"

                    # Special case: if this is a SOF PID, we'll extract
                    # the frame number from this, rather than our typical
                    # token fields.
                    with m.If(current_pid == self.SOF_PID):
                        m.d.usb += [
                            self.interface.frame      .eq(token_data),
                            self.interface.new_frame  .eq(1),
                        ]

                    # Otherwise, extract the address and endpoint from the token,
                    # and report the captured pid.
                    with m.Else():

                        # If we're filtering by address, only count this token if it's releveant to our address.
                        # Otherwise, always count tokens -- we'll report the address on the output.
                        token_applicable = (token_data[0:7] == self.address) if self.filter_by_address else True
                        with m.If(token_applicable):
                            m.d.usb += [
                                self.interface.pid        .eq(current_pid),
                                self.interface.new_token  .eq(1),

                                Cat(self.interface.address, self.interface.endpoint).eq(token_data)
                            ]

                            # Start our interpacket-delay timer.
                            m.d.comb += timer.start.eq(1)

                        # If we don't count the token, clear the state so we don't act on following packets.
                        with m.Else():
                            m.d.usb += self.interface.pid.eq(0)


                # Otherwise, if we get more data, we've received a malformed
                # token -- which we'll ignore.
                with m.Elif(self.utmi.rx_valid):
                    m.next="IRRELEVANT"


            # NON_TOKEN -- we've encountered a non-token packet; wait for it to end
            with m.State("IRRELEVANT"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

        return m



class USBHandshakeDetector(Elaboratable):
    """ Gateware that detects handshake packets.

    Attributes
    -----------
    detected: HandshakeExchangeInterface
        Strobes that indicate which handshakes we're detecting.

    Parameters
    ----------
    utmi: [UTMIInterface, UTMITranslator]
        The UTMI interface to listen on.
    """

    ACK_PID   = 0b0010
    NAK_PID   = 0b1010
    STALL_PID = 0b1110
    NYET_PID  = 0b0110

    def __init__(self, *, utmi):
        self.utmi = utmi

        #
        # I/O port
        #
        self.detected = HandshakeExchangeInterface(is_detector=True)


    def elaborate(self, platform):
        m = Module()

        active_pid = Signal(4)

        # Keep our strobes un-asserted unless otherwise specified.
        m.d.usb += [
            self.detected.ack    .eq(0),
            self.detected.nak    .eq(0),
            self.detected.stall  .eq(0),
            self.detected.nyet   .eq(0),
        ]


        with m.FSM(domain="usb"):

            # IDLE -- waiting for a packet to be presented
            with m.State("IDLE"):
                with m.If(self.utmi.rx_active):
                    m.next = "READ_PID"


            # READ_PID -- read the packet's ID.
            with m.State("READ_PID"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                with m.Elif(self.utmi.rx_valid):
                    is_valid_pid = (self.utmi.rx_data[0:4] == ~self.utmi.rx_data[4:8])

                    # If we have a valid PID, move to capture it.
                    with m.If(is_valid_pid):
                        m.d.usb += active_pid.eq(self.utmi.rx_data)
                        m.next = "AWAIT_COMPLETION"

                    # Otherwise, ignore this packet as a non-token.
                    with m.Else():
                        m.next = "IRRELEVANT"


            # TOKEN_COMPLETE: we've received a full token; and now need to wait
            # for the packet to be complete.
            with m.State("AWAIT_COMPLETION"):

                # Once our receive is complete, we can parse the PID
                # and identify the event.
                with m.If(~self.utmi.rx_active):
                    m.d.usb += [
                        self.detected.ack    .eq(active_pid == self.ACK_PID),
                        self.detected.nak    .eq(active_pid == self.NAK_PID),
                        self.detected.stall  .eq(active_pid == self.STALL_PID),
                        self.detected.nyet   .eq(active_pid == self.NYET_PID),
                    ]
                    m.next="IDLE"

                # Otherwise, if we get more data, this isn't a valid handshake.
                # Skip this packet as irrelevant.
                with m.Elif(self.utmi.rx_valid):
                    m.next="IRRELEVANT"


            # IRRELEVANT -- we've encountered a malformed or non-handshake packet
            with m.State("IRRELEVANT"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

        return m




class USBDataPacketCRC(Elaboratable):
    """ Gateware that computes a running CRC-16.

    By default, this module has no connections to the modules that use it.

    These are added using :attr:`add_interface`; this module supports an arbitrary
    number of connection interfaces; see :attr:`add_interface()` for restrictions.

    Attributes
    ----------
    rx_data: Signal(8), input
        Receive data input; can be carried directly from a UTMI interface.
    rx_valid: Signal(), input
        Receive validity signal; can be carried directly from a UTMI interface.

    tx_data: Signal(8), input
        Transmit data input; can be carried directly from a UTMI interface.
    tx_valid: Signal(), input
        When high, the `tx_data` input is used to update the CRC.

    Parameters
    ----------
    initial_value: [int, Const]
            The initial value of the CRC shift register; the USB default is used if not provided.
    """

    def __init__(self, initial_value=0xFFFF):

        self._initial_value = initial_value

        # List of interfaces to work with.
        # This list is populated dynamically by calling .add_interface().
        self._interfaces    = []

        #
        # I/O port
        #
        self.clear = Signal()

        self.rx_data  = Signal(8)
        self.rx_valid = Signal()

        self.tx_data  = Signal(8)
        self.tx_valid = Signal()

        self.crc   = Signal(16, init=initial_value)


    def add_interface(self, interface : DataCRCInterface):
        """ Adds an interface to the CRC generator module.

        Each interface can reset the CRC; and can read the current CRC value.
        No arbitration is performed; it's assumed that no more than one interface
        will be computing a running CRC at at time.

        Parameters
        ----------
        interface: DataCRCInterface
            The interface to be added; accepts control signals from other modules, and
            brings CRC output to them. This method can be called multiple times to generate
            multiplpe CRCs.
        """
        self._interfaces.append(interface)


    def _generate_next_crc(self, current_crc, data_in):
        """ Generates the next round of a bytewise USB CRC16. """
        xor_reduce = lambda bits : functools.reduce(operator.__xor__, bits)

        # Extracted from the USB spec's definition of the CRC16 polynomial.
        return Cat(
            xor_reduce(data_in)      ^ xor_reduce(current_crc[ 8:16]),
            xor_reduce(data_in[0:7]) ^ xor_reduce(current_crc[ 9:16]),
            xor_reduce(data_in[6:8]) ^ xor_reduce(current_crc[ 8:10]),
            xor_reduce(data_in[5:7]) ^ xor_reduce(current_crc[ 9:11]),
            xor_reduce(data_in[4:6]) ^ xor_reduce(current_crc[10:12]),
            xor_reduce(data_in[3:5]) ^ xor_reduce(current_crc[11:13]),
            xor_reduce(data_in[2:4]) ^ xor_reduce(current_crc[12:14]),
            xor_reduce(data_in[1:3]) ^ xor_reduce(current_crc[13:15]),

            xor_reduce(data_in[0:2]) ^ xor_reduce(current_crc[14:16]) ^ current_crc[0],
            data_in[0] ^ current_crc[1] ^ current_crc[15],
            current_crc[2],
            current_crc[3],
            current_crc[4],
            current_crc[5],
            current_crc[6],
            xor_reduce(data_in) ^ xor_reduce(current_crc[7:16]),
        )


    def elaborate(self, platform):
        m = Module()

        # Register that contains the running CRCs.
        crc        = Signal(16, init=self._initial_value)

        # Signal that contains the output version of our active CRC.
        output_crc = Signal.like(crc)

        # We'll clear our CRC whenever any of our interfaces request it.
        start_signals = (interface.start for interface in self._interfaces)
        clear = functools.reduce(operator.__or__, start_signals)

        # If we're clearing our CRC in progress, move our holding register back to
        # our initial value.
        with m.If(clear):
            m.d.usb += crc.eq(self._initial_value)

        # Otherwise, update the CRC whenever we have new data.
        with m.Elif(self.rx_valid):
            m.d.usb += crc.eq(self._generate_next_crc(crc, self.rx_data))
        with m.Elif(self.tx_valid):
            m.d.usb += crc.eq(self._generate_next_crc(crc, self.tx_data))

        # Convert from our intermediary "running CRC" format into the current CRC-16...
        m.d.comb += output_crc.eq(~crc[::-1])

        # ... and connect it to each of our interfaces.
        for interface in self._interfaces:
            m.d.comb += interface.crc.eq(output_crc)

        return m


class USBDataPacketReceiver(Elaboratable):
    """ Gateware that converts received USB data packets into a data-stream packets.

    It's important to note that packet payloads are mostly directly carried over from UTMI.
    Since USB data is received -prior- to its CRC, one cannot know if a packet is valid until
    after it has been compeltely received. As a result, this interface will generate data of
    unknown validity, followed by a strobe on either :attr:`packet_complete` or :attr:`crc_mismatch`.
    The receiving interface must be prepared to handle :attr:`crc_mismatch` by discarding the received
    data.


    Attributes
    ----------
    data_crc: DataCRCInterface
        Connection to the CRC generator.
    timer: InterpacketTimerInterface
        Connection to our interpacket timer.
    stream: USBOutDataStream, output
        Stream that carries captured packet data.

    active_pid: Signal(4), output
        The PID of the data currently being received.
    packet_id: Signal(4), output
        The packet ID of the most recently captured PID. Becomes valid simultaneous to a strobe on
        :attr:`packet_complete` or :attr:`crc_mismatch`.

    packet_complete: Signal(), output
        Strobe that pulses high when a new packet is delivered with a valid CRC.
    crc_mismatch: Signal(), output
        Strobe that pulses high when the given packet has a CRC mismatch; and thus the data
        received this far should be discarded.
    ready_for_response: Signal(), output
        Strobe that indicates that an inter-packet delay has passed since :attr:`packet_complete`,
        and thus we're now ready to respond with a handshake.

    Parameters
    ----------
    utmi: UTMIInterface, or equivalent
        The UTMI bus to observe.
    max_packet_size: int
        The maximum packet (payload) size to be deserialized, in bytes.

    standalone: bool
        Debug value. If True, a submodule CRC generator will be created.
    speed: USBSpeed
        USBSpeed signal or constant that specifies our speed in standalone mode.
    """

    _DATA_SUFFIX = 0b11

    def __init__(self, *, utmi, standalone=False, speed=None):

        self.utmi        = utmi
        self.standalone  = standalone
        self.speed       = speed

        #
        # I/O port
        #
        self.data_crc           = DataCRCInterface()
        self.timer              = InterpacketTimerInterface()
        self.stream             = USBOutStreamInterface()

        self.active_pid         = Signal(4)

        self.packet_complete    = Signal()
        self.ready_for_response = Signal()
        self.crc_mismatch       = Signal()
        self.packet_id          = Signal(4)


    def elaborate(self, platform):
        m = Module()

        # If we're in standalone mode, create our dependencies for us.
        if self.standalone:
            m.submodules.crc = crc = USBDataPacketCRC()
            crc.add_interface(self.data_crc)

            m.submodules.timer = timer = USBInterpacketTimer()
            timer.add_interface(self.timer)

            if not self.speed:
                self.speed = USBSpeed.FULL

            m.d.comb += [

                # Connect our CRC generator...
                crc.rx_data           .eq(self.utmi.rx_data),
                crc.rx_valid          .eq(self.utmi.rx_valid),
                crc.tx_valid          .eq(0),

                # ... and our timer.
                timer.speed           .eq(self.speed)
            ]


        # CRC-16 tracking signals.
        last_byte_crc = Signal(16)
        last_word_crc = Signal(16)

        # Keeps track of the most recently received word; for CRC comparison/removal.
        data_pipeline     = Signal(8 * 2)

        # Keep our control signals + strobes un-asserted unless otherwise specified.
        m.d.usb  += [
            self.packet_complete .eq(0),
            self.crc_mismatch    .eq(0),
        ]
        m.d.comb += [
            self.stream.next     .eq(0),
            self.data_crc.start  .eq(0),
        ]


        with m.FSM(domain="usb"):

            # IDLE -- waiting for a packet to be presented
            with m.State("IDLE"):

                with m.If(self.utmi.rx_active):
                    m.next = "READ_PID"

            # READ_PID -- read the packet's ID.
            with m.State("READ_PID"):

                # Clear our CRC; as we're potentially about to start a new packet.
                m.d.comb += self.data_crc.start.eq(1)

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                with m.Elif(self.utmi.rx_valid):
                    is_data      = (self.utmi.rx_data[0:2] == self._DATA_SUFFIX)
                    is_valid_pid = (self.utmi.rx_data[0:4] == ~self.utmi.rx_data[4:8])

                    # If this is a data packet, capture its PID.
                    with m.If(is_valid_pid & is_data):
                        m.d.usb += self.active_pid.eq(self.utmi.rx_data),
                        m.next = "RECEIVE_FIRST_BYTE"

                    # Otherwise, ignore this packet.
                    with m.Else():
                        m.next = "IRRELEVANT"


            # RECEIVE_FIRST_BYTE -- capture the first byte into our pipeline.
            # We'll always pipeline two bytes before we start emitting; as we won't want to
            # pass through the last two bytes (the CRC).
            with m.State("RECEIVE_FIRST_BYTE"):

                with m.If(self.utmi.rx_valid):
                    m.d.usb += [
                        data_pipeline[8:]  .eq(self.utmi.rx_data),
                        last_byte_crc       .eq(self.data_crc.crc)
                    ]
                    m.next = 'RECEIVE_SECOND_BYTE'

                # If our packet stops before we see the first to bytes, we'll return to idle.
                # There's nothing to clean up, as we've never touched the stream.
                with m.If(~self.utmi.rx_active):
                    m.next = 'IDLE'


            # RECEIVE_SECOND_BYTE-- capture the second byte into our pipeline.
            with m.State("RECEIVE_SECOND_BYTE"):

                with m.If(self.utmi.rx_valid):
                    m.d.usb += [
                        data_pipeline[8:]   .eq(self.utmi.rx_data),
                        data_pipeline[0:8]  .eq(data_pipeline[8:]),

                        last_byte_crc       .eq(self.data_crc.crc),
                        last_word_crc       .eq(last_byte_crc),
                    ]
                    m.next = 'RECEIVE_AND_EMIT'

                # If our packet stops before we see the first to bytes, we'll return to idle.
                # There's nothing to clean up, as we've never touched the stream.
                with m.Elif(~self.utmi.rx_active):
                    m.next = 'IDLE'


            # RECEIVE_AND_EMIT -- receive bytes into our pipeline, and emit them.
            # Now that we have more than two bytes captured, we can start emitting bytes.
            # We'll always be emitting bytes that are two old -- so we can stop before our CRC.:
            with m.State("RECEIVE_AND_EMIT"):
                m.d.comb += self.stream.valid.eq(1)

                with m.If(self.utmi.rx_valid):

                    m.d.comb += [
                        # Emit the current packet...
                        self.stream.payload  .eq(data_pipeline[0:8]),
                        self.stream.next     .eq(1),
                    ]

                    m.d.usb += [

                        # ... capture the incoming one...
                        data_pipeline[8:]   .eq(self.utmi.rx_data),
                        data_pipeline[0:8]  .eq(data_pipeline[8:]),

                        # ... and update our cached CRCs.
                        last_byte_crc       .eq(self.data_crc.crc),
                        last_word_crc       .eq(last_byte_crc),
                    ]


                # Once we stop receiving data, check our CRC and finish.
                with m.If(~self.utmi.rx_active):

                    # If our CRC matches, this is a valid packet!
                    with m.If(last_word_crc == data_pipeline):

                        # Indicate so...
                        m.d.usb += [
                            self.packet_id       .eq(self.active_pid),
                            self.packet_complete .eq(1)
                        ]

                        # ... start counting our interpacket delay...
                        m.d.comb += [
                            self.timer.start  .eq(1)
                        ]

                        # ... and wait for it to complete.
                        m.next = 'INTERPACKET_DELAY'


                    # Otherwise, flag this as a CRC mismatch.
                    with m.Else():
                        m.d.usb += [
                            self.crc_mismatch    .eq(1)
                        ]

                        # ... and return to IDLE.
                        m.next = "IDLE"


            # INTERPACKET_DELAY -- we've received a valid packet; wait for an
            # interpacket delay before moving back to IDLE.
            with m.State("INTERPACKET_DELAY"):

                with m.If(self.timer.tx_allowed):
                    m.d.comb += self.ready_for_response.eq(1)
                    m.next = 'IDLE'


            # IRRELEVANT -- we've encountered a malformed or non-DATA packet.
            with m.State("IRRELEVANT"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

        return m



class USBDataPacketDeserializer(Elaboratable):
    """ Gateware that captures USB data packet contents and parallelizes them.

    Attributes
    ----------
    data_crc: DataCRCInterface
        Connection to the CRC generator.

    new_packet: Signal(), output
        Strobe that pulses high for a single cycle when a new packet is delivered.
    packet_id: Signal(4), output
        The packet ID of the captured PID.

    packet: Signal(max_packet_size), output
        Packet data for a the most recently received packet.
    length: Signal(range(0, max_packet_length +1)), output
        The length of the packet data presented on the packet[] output.

    Parameters
    ----------
    utmi: UTMIInterface, or equivalent
        The UTMI bus to observe.
    max_packet_size: int
        The maximum packet (payload) size to be deserialized, in bytes.
    create_crc_generator: bool
        If True, a submodule CRC generator will be created. Excellent for testing.
    """

    _DATA_SUFFIX = 0b11

    def __init__(self, *, utmi, max_packet_size=64, create_crc_generator=False):

        self.utmi                 = utmi
        self._max_packet_size     = max_packet_size
        self.create_crc_generator = create_crc_generator

        #
        # I/O port
        #
        self.data_crc    = DataCRCInterface()

        self.new_packet  = Signal()

        self.packet_id   = Signal(4)
        self.packet      = Array(Signal(8, name=f"packet_{i}") for i in range(max_packet_size))
        self.length      = Signal(range(0, max_packet_size + 1))


    def elaborate(self, platform):
        m = Module()

        max_size_with_crc = self._max_packet_size + 2

        # If we're creating an internal CRC generator, create a submodule
        # and hook it up.
        if self.create_crc_generator:
            m.submodules.crc = crc = USBDataPacketCRC()
            crc.add_interface(self.data_crc)

            m.d.comb += [
                crc.rx_data           .eq(self.utmi.rx_data),
                crc.rx_valid          .eq(self.utmi.rx_valid),
                crc.tx_valid           .eq(0)
            ]

        # CRC-16 tracking signals.
        last_byte_crc = Signal(16)
        last_word_crc = Signal(16)

        # Currently captured PID.
        active_pid         = Signal(4)

        # Active packet transfer.
        active_packet      = Array(Signal(8) for _ in range(max_size_with_crc))
        position_in_packet = Signal(range(0, max_size_with_crc))

        # Keeps track of the most recently received word; for CRC comparison.
        last_word          = Signal(16)

        # Keep our control signals + strobes un-asserted unless otherwise specified.
        m.d.usb += self.new_packet      .eq(0)
        m.d.comb += self.data_crc.start  .eq(0)

        with m.FSM(domain="usb"):

            # IDLE -- waiting for a packet to be presented
            with m.State("IDLE"):

                with m.If(self.utmi.rx_active):
                    m.next = "READ_PID"

            # READ_PID -- read the packet's ID.
            with m.State("READ_PID"):
                # Clear our CRC; as we're potentially about to start a new packet.
                m.d.comb += self.data_crc.start.eq(1)

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

                with m.Elif(self.utmi.rx_valid):
                    is_data      = (self.utmi.rx_data[0:2] == self._DATA_SUFFIX)
                    is_valid_pid = (self.utmi.rx_data[0:4] == ~self.utmi.rx_data[4:8])

                    # If this is a data packet, capture it.
                    with m.If(is_valid_pid & is_data):
                        m.d.usb += [
                            active_pid          .eq(self.utmi.rx_data),
                            position_in_packet  .eq(0)
                        ]
                        m.next = "CAPTURE_DATA"

                    # Otherwise, ignore this packet.
                    with m.Else():
                        m.next = "IRRELEVANT"


            with m.State("CAPTURE_DATA"):

                # If we have a new byte of data, capture it.
                with m.If(self.utmi.rx_valid):

                    # If this would over-fill our internal buffer, fail out.
                    with m.If(position_in_packet >= max_size_with_crc):
                        # TODO: potentially signal the babble?
                        m.next = "IRRELEVANT"

                    with m.Else():
                        m.d.usb += [
                            active_packet[position_in_packet]  .eq(self.utmi.rx_data),
                            position_in_packet                 .eq(position_in_packet + 1),

                            last_word     .eq(Cat(last_word[8:], self.utmi.rx_data)),

                            last_word_crc .eq(last_byte_crc),
                            last_byte_crc .eq(self.data_crc.crc),
                        ]


                # If this is the end of our packet, validate our CRC and finish.
                with m.If(~self.utmi.rx_active):

                    with m.If(last_word_crc == last_word):
                        m.d.usb += [
                            self.packet_id   .eq(active_pid),
                            self.length      .eq(position_in_packet - 2),
                            self.new_packet  .eq(1)
                        ]

                        for i in range(self._max_packet_size):
                            m.d.usb += self.packet[i].eq(active_packet[i]),

                        m.next = "IDLE"

            # IRRELEVANT -- we've encountered a malformed or non-handshake packet
            with m.State("IRRELEVANT"):

                with m.If(~self.utmi.rx_active):
                    m.next = "IDLE"

        return m


class USBDataPacketGenerator(Elaboratable):
    """ Module that converts a FIFO-style stream into a USB data packet.

    Handles steps such as PID generation and CRC-16 injection.

    As a special case, if the stream pulses `last` (with valid=1) without pulsing
    `first`, we'll send a zero-length packet.

    Attributes
    ----------

    data_pid: Signal(2), input
        The data packet number to use. The potential PIDS are: 0 = DATA0, 1 = DATA1,
        2 = DATA2, 3 = MDATA; the interface is designed so that most endpoints can tie the MSb to
        zero and then perform PID toggling by toggling the LSb.

    crc: DataCRCInterface
        Interface to our data CRC generator.
    stream: USBInStreamInterface
        Stream input for the raw data to be transmitted.
    tx: UTMITransmitInterface
        UTMI-subset transmit interface

    Parameters
    ----------
    standalone: bool
        If True, this unit will include its internal CRC generator. Perfect for unit testing or debugging.
    """

    def __init__(self, standalone=False):

        self.standalone = standalone

        #
        # I/O port
        #
        self.data_pid     = Signal(2)

        self.crc          = DataCRCInterface()
        self.stream       = USBInStreamInterface()
        self.tx           = UTMITransmitInterface()


    def elaborate(self, platform):
        m = Module()

        # Create a mux that maps our data_pid value to our actual data PID.
        data_pids = Array([
            Const(0xC3, shape=8), # DATA0
            Const(0x4B, shape=8), # DATA1
            Const(0x87, shape=8), # DATA2
            Const(0x0F, shape=8)  # DATAM
        ])

        # Stores the current data pid; latched in at the start of a transmission.
        current_data_pid = Signal(8)

        # Register that stores the final CRC byte.
        # Capturing this before the end of the packet ensures we can still send
        # the correct final CRC byte; even if the CRC generator updates its computation
        # when the first byte of the CRC is transmitted.
        remaining_crc = Signal(8)

        # Flag that stores whether we're sending a zero-length packet.
        is_zlp = Signal()

        # If we're creating an internal CRC generator, create a submodule
        # and hook it up.
        if self.standalone:
            m.submodules.crc = crc = USBDataPacketCRC()
            crc.add_interface(self.crc)

            m.d.comb += [
                crc.rx_valid          .eq(0),

                crc.tx_data           .eq(self.stream.payload),
                crc.tx_valid          .eq(self.tx.ready)
            ]

        with m.FSM(domain="usb"):

            # IDLE -- waiting for an active transmission to start.
            with m.State('IDLE'):

                # We won't consume any data while we're in the IDLE state.
                m.d.comb += self.stream.ready.eq(0)

                # Latch in the requested data PID.
                m.d.usb += current_data_pid.eq(data_pids[self.data_pid])

                # Once a packet starts, we'll need to transmit the data PID.
                with m.If(self.stream.first & self.stream.valid):
                    m.d.usb += is_zlp.eq(0)
                    m.next = "SEND_PID"

                # Special case: if `last` pulses without first, we'll consider this
                # a zero-length packet ("a packet without a first byte").
                with m.Elif(self.stream.last & self.stream.valid):
                    m.d.usb += is_zlp.eq(1)
                    m.next = "SEND_PID"


            # SEND_PID -- prepare for the transaction by sending the data packet ID.
            with m.State('SEND_PID'):

                m.d.comb += [
                    # Prepare for a new payload by starting a new CRC calculation.
                    self.crc.start     .eq(1),

                    # Send the USB packet ID for our data packet...
                    self.tx.data       .eq(current_data_pid),
                    self.tx.valid      .eq(1),

                    # ... and don't consume any data.
                    self.stream.ready  .eq(0)
                ]

                # Advance once the PHY accepts our PID.
                with m.If(self.tx.ready):

                    # If this is a ZLP, we don't have a payload to send.
                    # Skip directly to sending our CRC.
                    with m.If(is_zlp):
                        m.next = 'SEND_CRC_FIRST'

                    # Otherwise, we have a payload. Send it.
                    with m.Else():
                        m.next = 'SEND_PAYLOAD'


            # SEND_PAYLOAD -- send the data payload for our stream
            with m.State('SEND_PAYLOAD'):

                # While sending the payload, we'll essentially connect
                # our stream directly through to the ULPI transmitter.
                m.d.comb += self.stream.bridge_to(self.tx)

                # We'll stop sending once the packet ends, and move on to our CRC.
                with m.If(self.tx.ready & (self.stream.last | ~self.stream.valid)):
                    m.next = 'SEND_CRC_FIRST'


            # SEND_CRC_FIRST -- send the first byte of the packet's CRC
            with m.State('SEND_CRC_FIRST'):

                # Capture the current CRC for use in the next byte...
                m.d.usb += remaining_crc.eq(self.crc.crc[8:])

                # Send the relevant CRC byte...
                m.d.comb += [
                    self.tx.data       .eq(self.crc.crc[0:8]),
                    self.tx.valid      .eq(1),
                ]

                # ... and move on to the next one.
                with m.If(self.tx.ready):
                    m.next = 'SEND_CRC_SECOND'


            # SEND_CRC_LAST -- send the last byte of the packet's CRC
            with m.State('SEND_CRC_SECOND'):

                # Send the relevant CRC byte...
                m.d.comb += [
                    self.tx.data       .eq(remaining_crc),
                    self.tx.valid      .eq(1),
                ]

                # ... and return to idle.
                with m.If(self.tx.ready):
                    m.next = 'IDLE'

        return m



class USBHandshakeGenerator(Elaboratable):
    """ Module that generates handshake packets, on request.

    Attributes:

    issue_ack: Signal(), input
        Pulsed to generate an ACK handshake packet.
    issue_nak: Signal(), input
        Pulsed to generate a NAK handshake packet.
    issue_stall: Signal(), input
        Pulsed to generate a STALL handshake.

    tx: UTMITransmitInterface
        Interface to the relevant UTMI interface.
    """

    # Full contents of an ACK, NAK, and STALL packet.
    # These include the four check bits; which consist of the inverted PID.
    _PACKET_ACK   = 0b11010010
    _PACKET_NAK   = 0b01011010
    _PACKET_STALL = 0b00011110

    def __init__(self):

        #
        # I/O port
        #
        self.issue_ack    = Signal()
        self.issue_nak    = Signal()
        self.issue_stall  = Signal()

        self.tx           = UTMITransmitInterface()


    def elaborate(self, platform):
        m = Module()

        with m.FSM(domain="usb"):

            # IDLE -- we haven't yet received a request to transmit
            with m.State('IDLE'):
                m.d.comb += self.tx.valid.eq(0)

                # Wait until we have an ACK, NAK, or STALL request;
                # Then set our data value to the appropriate PID,
                # in preparation for the next cycle.

                with m.If(self.issue_ack):
                    m.d.usb += self.tx.data  .eq(self._PACKET_ACK),
                    m.next = 'TRANSMIT'

                with m.If(self.issue_nak):
                    m.d.usb += self.tx.data  .eq(self._PACKET_NAK),
                    m.next = 'TRANSMIT'

                with m.If(self.issue_stall):
                    m.d.usb += self.tx.data  .eq(self._PACKET_STALL),
                    m.next = 'TRANSMIT'


            # TRANSMIT -- send the handshake.
            with m.State('TRANSMIT'):
                m.d.comb += self.tx.valid.eq(1)

                # Once we know the transmission will be accepted, we're done!
                # Move back to IDLE.
                with m.If(self.tx.ready):
                    m.next = 'IDLE'

        return m


class USBInterpacketTimer(Elaboratable):
    """ Module that tracks inter-packet timings, enforcing spec-mandated packet gaps.

    Ports other than :attr:`speed` are added dynamically via :method:add_interface`.

    Attributes
    ----------
    speed: Signal(2), input
        The device's current operating speed. Should be a USBSpeed enumeration value --
        0 for high, 1 for full, 2 for low.

    """

    # Per the USB 2.0 and ULPI 1.1 specifications, after receipt:
    #   - A FS/LS device needs to wait 2 bit periods before transmitting; and must
    #     respond before 6.5 bit times pass. [USB2, 7.1.18.1]
    #   - Two FS bit periods is equivalent to 10 ULPI clocks, and two LS periods is
    #     equivalent to 80 ULPI clocks. 6.5 FS bit periods is equivalent to 32 ULPI clocks,
    #     and 6.5 LS bit periods is equivalent to 260 ULPI clocks. [ULPI 1.1, Figure 18].
    #   - A HS device needs to wait 8 HS bit periods before transmitting [USB2, 7.1.18.2].
    #     Each ULPI cycle is 8 HS bit periods, so we'll only need to wait one cycle.
    _HS_RX_TO_TX_DELAY     = {60e6: (  1,  24)}
    _FS_RX_TO_TX_DELAY     = {60e6: ( 10,  32), 12e6: (2, 7)}
    _LS_RX_TO_TX_DELAY     = {60e6: ( 80, 260)}

    # Per the USB 2.0 and ULPI 1.1 specifications, after transission:
    #   - A FS/LS can assume it won't receive a response after 16 bit times [USB2, 7.1.18.1].
    #     This is equivalent to 80 ULPI clocks (FS), or 640 ULPI clocks (LS).
    #   - A HS device can assume it won't receive a response after 736 bit times.
    #     This is equivalent to 92 ULPI clocks.
    _HS_TX_TO_RX_TIMEOUT = {60e6:  92}
    _FS_TX_TO_RX_TIMEOUT = {60e6:  80, 12e6: 16}
    _LS_TX_TO_RX_TIMEOUT = {60e6: 640}


    def __init__(self, domain_clock=60e6, fs_only=False):
        self._fs_only = fs_only

        # Start off with empty delays -- this doesn't change anything, but makes
        # linters happy. :)
        self._hs_rx_to_tx_delay   = None
        self._ls_rx_to_tx_delay   = None
        self._hs_rx_to_tx_timeout = None
        self._ls_rx_to_tx_timeout = None

        # Validate that we have a usable FS Rx/Tx delay.
        if domain_clock not in self._FS_RX_TO_TX_DELAY:
            raise ValueError(f"Domain clock must be in {self._FS_TX_TO_RX_TIMEOUT.keys()}, not {domain_clock}.")

        # Capture our FS delay for the current clock speed.
        self._fs_rx_to_tx_delay   = self._FS_RX_TO_TX_DELAY[domain_clock]
        self._fs_tx_to_rx_timeout = self._FS_TX_TO_RX_TIMEOUT[domain_clock]
        self._counter_max = self._FS_TX_TO_RX_TIMEOUT[domain_clock]

        # If we're not in a FS-only configuration, capture our other delays.
        if not self._fs_only:
            if domain_clock not in self._HS_RX_TO_TX_DELAY:
                raise ValueError(f"Domain clock must be in {self._FS_TX_TO_RX_TIMEOUT.keys()}, not {domain_clock}.")

            # Capute our HS and LS delays for the given clock speed.
            self._hs_rx_to_tx_delay   = self._HS_RX_TO_TX_DELAY[domain_clock]
            self._ls_rx_to_tx_delay   = self._LS_RX_TO_TX_DELAY[domain_clock]
            self._hs_tx_to_rx_timeout = self._HS_TX_TO_RX_TIMEOUT[domain_clock]
            self._ls_tx_to_rx_timeout = self._LS_TX_TO_RX_TIMEOUT[domain_clock]
            self._counter_max         = self._LS_TX_TO_RX_TIMEOUT[domain_clock]



        # List of interfaces to users of this module.
        self._interfaces               = []

        #
        # I/O port
        #
        self.speed = Signal(2)


    def add_interface(self, interface: InterpacketTimerInterface):
        """ Adds a connection to a user of this module.

        This module performs no multiplexing; it's assumed only one interface will be active at a time.

        Parameters
        ---------
        interface: InterpacketTimerInterface
            The InterPacketTimer interface to add to our module.
        """
        self._interfaces.append(interface)


    def elaborate(self, platform):
        m = Module()

        # Internal signals representing each of our timeouts.
        rx_to_tx_at_min  = Signal()
        rx_to_tx_at_max  = Signal()
        tx_to_rx_timeout = Signal()

        # Create a counter that will track our interpacket delays.
        # This should be able to count up to our longest delay. We'll allow our
        # counter to be able to increment one past its maximum, and let it saturate
        # there, after the count.
        counter = Signal(range(0, self._counter_max + 2))

        # Reset our timer whenever any of our interfaces request a timer start.
        reset_signals = (interface.start for interface in self._interfaces)
        any_reset = functools.reduce(operator.__or__, reset_signals)

        # When a reset is requested, start the counter from 0.
        with m.If(any_reset):
            m.d.usb += counter.eq(0)
        with m.Elif(counter < self._counter_max + 1):
            m.d.usb += counter.eq(counter + 1)

        #
        # Create our counter-progress strobes.
        # This could be made less repetitive, but spreading it out here
        # makes the documentation above clearer.
        #
        with m.If(self.speed == USBSpeed.HIGH):
            if not self._fs_only:
                m.d.comb += [
                    rx_to_tx_at_min   .eq(counter == self._hs_rx_to_tx_delay[0]),
                    rx_to_tx_at_max   .eq(counter == self._hs_rx_to_tx_delay[1]),
                    tx_to_rx_timeout  .eq(counter == self._hs_tx_to_rx_timeout)
                ]
        with m.Elif(self.speed == USBSpeed.FULL):
            m.d.comb += [
                rx_to_tx_at_min   .eq(counter == self._fs_rx_to_tx_delay[0]),
                rx_to_tx_at_max   .eq(counter == self._fs_rx_to_tx_delay[1]),
                tx_to_rx_timeout  .eq(counter == self._fs_tx_to_rx_timeout)
            ]
        with m.Else():
            if not self._fs_only:
                m.d.comb += [
                    rx_to_tx_at_min   .eq(counter == self._hs_rx_to_tx_delay[0]),
                    rx_to_tx_at_max   .eq(counter == self._hs_rx_to_tx_delay[1]),
                    tx_to_rx_timeout  .eq(counter == self._hs_tx_to_rx_timeout)
                ]

        # Tie our strobes to each of our consumers.
        for interface in self._interfaces:
            m.d.comb += [
                interface.tx_allowed.eq(rx_to_tx_at_min),
                interface.tx_timeout.eq(rx_to_tx_at_max),
                interface.rx_timeout.eq(tx_to_rx_timeout)
            ]


        return m




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Header Packet data interfacing definitions."""

import operator
import functools

from enum import IntEnum

from amaranth           import *
from amaranth.hdl.rec   import Layout

from ....stream.arbiter import StreamArbiter


class HeaderPacket(Record):
    """ Container that represents a Header Packet. """

    # Create overrideable constants that allow us to specialize
    # the data words of our headers in subclasses.
    DW0_LAYOUT = [('dw0', 32)]
    DW1_LAYOUT = [('dw1', 32)]
    DW2_LAYOUT = [('dw2', 32)]

    LINK_LAYER_FIELDS = [
        ('crc16',           16),
        ('sequence_number',  3),
        ('dw3_reserved',     3),
        ('hub_depth',        3),
        ('delayed',          1),
        ('deferred',         1),
        ('crc5',             5),
    ]

    def get_type(self):
        """ Returns the selection of bits in DW0 that encode the packet type. """
        return self.dw0[0:5]


    @classmethod
    def get_layout(cls):
        """ Computes the layout for the HeaderPacket (sub)class. """
        return [
            *cls.DW0_LAYOUT,
            *cls.DW1_LAYOUT,
            *cls.DW2_LAYOUT,
            *cls.LINK_LAYER_FIELDS
        ]


    def __init__(self):
        super().__init__(self.get_layout(), name=self.__class__.__name__)



class HeaderQueue(Record):
    """ Record representing a header, and stream-link control signals.

    Attributes
    ----------
    valid: Signal(), producer to consumer
        Indicates that the data in :attr:``header`` is valid and ready to be consumed.
    header: HeaderPacket(), producer to consumer
        Contains a full set of header packet data.
    ready: Signal(), consumer to producer
        Strobed by the consumer to indicate that it has accepted the given header.
    """

    def __init__(self, *, header_type=HeaderPacket):
        super().__init__([
            ('valid', 1),
            ('header', header_type.get_layout()),
            ('ready', 1),
        ], name="HeaderQueue")


    def get_type(self):
        """ Returns the selection of bits in the current header's that encode the packet type. """
        return self.header.dw0[0:5]


    def header_eq(self, other):
        """ Connects a producer (self) up to a consumer. """
        return [
            self.valid   .eq(other.valid),
            self.header  .eq(other.header),
            other.ready  .eq(self.ready)
        ]


    def stream_eq(self, other):
        """ Alias for ``header_eq`` that ensures we share a stream interface. """
        return self.header_eq(other)



class HeaderQueueArbiter(StreamArbiter):
    """ Gateware that accepts a collection of header queues, and merges them into a single queue.

    Add produces using ``add_producer``.

    Attributes
    ----------
    source: HeaderQueue(), output queue
        A single header queue that carries data from all producer queues.
    """

    def __init__(self):
        super().__init__(stream_type=HeaderQueue, domain="ss")


    def add_producer(self, interface: HeaderQueue):
        """ Adds a HeaderQueue interface that will add packets into this mux. """
        self.add_stream(interface)



class HeaderQueueDemultiplexer(Elaboratable):
    """ Gateware that accepts a single Header Queue, and routes it to multiple modules.

    Assumes that each type of header is handled by a separate module, and thus no two inputs
    will assert :attr:``ready`` at the same time.

    Add consumers using ``add_consumer``.

    Attributes
    ----------
    sink: HeaderQueue(), input queue
        The single header queue to be distributed to all of our consumers.
    """

    def __init__(self):
        self._consumers = []

        #
        # I/O port
        #
        self.sink = HeaderQueue()


    def add_consumer(self, interface: HeaderQueue):
        """ Adds a HeaderQueue interface that will consume packets from this mux. """
        self._consumers.append(interface)


    def elaborate(self, platform):
        m = Module()

        # Share the ``valid`` signal and header itself with every consumer.
        for consumer in self._consumers:
            m.d.comb += [
                consumer.valid   .eq(self.sink.valid),
                consumer.header  .eq(self.sink.header),
            ]

        # OR together all of the ``ready`` signals to produce our multiplex'd ready.
        sink_ready = functools.reduce(operator.__or__, (c.ready for c in self._consumers))
        m.d.comb += self.sink.ready.eq(sink_ready)


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test import usb_domain_test_case
from .test_usb2_packet  import USBPacketizerTest

from luna.gateware.usb.usb2         import USBSpeed
from luna.gateware.usb.usb2.request import USBSetupDecoder

class USBSetupDecoderTest(USBPacketizerTest):
    FRAGMENT_UNDER_TEST = USBSetupDecoder
    FRAGMENT_ARGUMENTS = {'standalone': True}


    def initialize_signals(self):

        # Assume high speed.
        yield self.dut.speed.eq(USBSpeed.HIGH)


    def provide_reference_setup_transaction(self):
        """ Provide a reference SETUP transaction. """

        # Provide our setup packet.
        yield from self.provide_packet(
            0b00101101, # PID: SETUP token.
            0b00000000, 0b00010000 # Address 0, endpoint 0, CRC
        )

        # Provide our data packet.
        yield from self.provide_packet(
            0b11000011,   # PID: DATA0
            0b0_10_00010, # out vendor request to endpoint
            12,           # request number 12
            0xcd, 0xab,   # value  0xABCD (little endian)
            0x23, 0x01,   # index  0x0123
            0x78, 0x56,   # length 0x5678
            0x3b, 0xa2,   # CRC
        )


    @usb_domain_test_case
    def test_valid_sequence_receive(self):
        dut = self.dut

        # Before we receive anything, we shouldn't have a new packet.
        self.assertEqual((yield dut.packet.received), 0)

        # Simulate the host sending basic setup data.
        yield from self.provide_reference_setup_transaction()

        # We're high speed, so we should be ACK'ing immediately.
        self.assertEqual((yield dut.ack), 1)

        # We now should have received a new setup request.
        yield
        self.assertEqual((yield dut.packet.received), 1)

        # Validate that its values are as we expect.
        self.assertEqual((yield dut.packet.is_in_request), 0       )
        self.assertEqual((yield dut.packet.type),          0b10    )
        self.assertEqual((yield dut.packet.recipient),     0b00010 )
        self.assertEqual((yield dut.packet.request),       12      )
        self.assertEqual((yield dut.packet.value),         0xabcd  )
        self.assertEqual((yield dut.packet.index),         0x0123  )
        self.assertEqual((yield dut.packet.length),        0x5678  )


    @usb_domain_test_case
    def test_fs_interpacket_delay(self):
        dut = self.dut

        # Place our DUT into full speed mode.
        yield dut.speed.eq(USBSpeed.FULL)

        # Before we receive anything, we shouldn't have a new packet.
        self.assertEqual((yield dut.packet.received), 0)

        # Simulate the host sending basic setup data.
        yield from self.provide_reference_setup_transaction()

        # We shouldn't ACK immediately; we'll need to wait our interpacket delay.
        yield
        self.assertEqual((yield dut.ack), 0)

        # After our minimum interpacket delay, we should see an ACK.
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.ack), 1)



    @usb_domain_test_case
    def test_short_setup_packet(self):
        dut = self.dut

        # Before we receive anything, we shouldn't have a new packet.
        self.assertEqual((yield dut.packet.received), 0)

        # Provide our setup packet.
        yield from self.provide_packet(
            0b00101101, # PID: SETUP token.
            0b00000000, 0b00010000 # Address 0, endpoint 0, CRC
        )

        # Provide our data packet; but shorter than expected.
        yield from self.provide_packet(
            0b11000011,                                     # PID: DATA0
            0b00100011, 0b01000101, 0b01100111, 0b10001001, # DATA
            0b00011100, 0b00001110                          # CRC
        )

        # This shouldn't count as a valid setup packet.
        yield
        self.assertEqual((yield dut.packet.received), 0)




#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test                   import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.link.receiver import RawHeaderPacketReceiver

class RawHeaderPacketReceiverTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = RawHeaderPacketReceiver

    def initialize_signals(self):
        yield self.dut.sink.valid.eq(1)

    def provide_data(self, *tuples):
        """ Provides the receiver with a sequence of (data, ctrl) values. """

        # Provide each word of our data to our receiver...
        for data, ctrl in tuples:
            yield self.dut.sink.data.eq(data)
            yield self.dut.sink.ctrl.eq(ctrl)
            yield


    @ss_domain_test_case
    def test_good_packet_receive(self):
        dut  = self.dut

        # Data input for an actual Link Management packet (seq #0).
        yield from self.provide_data(
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x00000280, 0b0000),
            (0x00010004, 0b0000),
            (0x00000000, 0b0000),
            (0x10001845, 0b0000),
        )

        # ... after a cycle to process, we should see an indication that the packet is good.
        yield from self.advance_cycles(2)
        self.assertEqual((yield dut.new_packet),   1)
        self.assertEqual((yield dut.bad_packet),   0)
        self.assertEqual((yield dut.bad_sequence), 0)


    @ss_domain_test_case
    def test_bad_sequence_receive(self):
        dut  = self.dut

        # Expect a sequence number other than the one we'll be providing.
        yield dut.expected_sequence.eq(3)

        # Data input for an actual Link Management packet (seq #0).
        yield from self.provide_data(
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x00000280, 0b0000),
            (0x00010004, 0b0000),
            (0x00000000, 0b0000),
            (0x10001845, 0b0000),
        )

        # ... after a cycle to process, we should see an indication that the packet is good.
        yield from self.advance_cycles(1)
        self.assertEqual((yield dut.new_packet),   0)
        self.assertEqual((yield dut.bad_packet),   0)
        self.assertEqual((yield dut.bad_sequence), 1)



    @ss_domain_test_case
    def test_bad_packet_receive(self):
        dut  = self.dut

        # Data input for an actual Link Management packet (seq #0),
        # but with the last word corrupted to invalidate our CRC16.
        yield from self.provide_data(
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x00000280, 0b0000),
            (0x00010004, 0b0000),
            (0xFFFFFFFF, 0b0000),
            (0x10001845, 0b0000),
        )

        # ... after a cycle to process, we should see an indication that the packet is bad.
        yield from self.advance_cycles(1)
        self.assertEqual((yield dut.new_packet),   0)
        self.assertEqual((yield dut.bad_packet),   1)
        self.assertEqual((yield dut.bad_sequence), 0)


    @ss_domain_test_case
    def test_bad_crc_and_sequence_receive(self):
        dut  = self.dut

        # Completely invalid link packet, guaranteed to have a bad sequence number & CRC.
        yield from self.provide_data(
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0xFFFFFFFF, 0b0000),
            (0xFFFFFFFF, 0b0000),
            (0xFFFFFFFF, 0b0000),
            (0xFFFFFFFF, 0b0000),
        )

        # Once we've processed this, we should see that there's a bad packet; but that it's
        # corrupted enough that our sequence no longer matters.
        yield from self.advance_cycles(1)
        self.assertEqual((yield dut.new_packet),   0)
        self.assertEqual((yield dut.bad_packet),   1)
        self.assertEqual((yield dut.bad_sequence), 0)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Core stream definitions. """

from amaranth         import Elaboratable, Signal, Module
from amaranth.hdl.rec import Record, DIR_FANIN, DIR_FANOUT


class StreamInterface(Record):
    """ Simple record implementing a unidirectional data stream.

    This class is similar to LiteX's streams; but instances may be optimized for
    interaction with USB PHYs. Accordingly, some uses may add restrictions; this
    is typically indicated by subclassing this interface.

    Attributes
    -----------
    valid: Signal(), from originator
        Indicates that the current payload bytes are valid pieces of the current transaction.
    first: Signal(), from originator
        Indicates that the payload byte is the first byte of a new packet.
    last: Signal(), from originator
        Indicates that the payload byte is the last byte of the current packet.
    payload: Signal(payload_width), from originator
        The data payload to be transmitted.

    ready: Signal(), from receiver
        Indicates that the receiver will accept the payload byte at the next active
        clock edge. Can be de-asserted to put backpressure on the transmitter.

    Parameters
    ----------
    payload_width: int
        The width of the stream's payload, in bits.
    extra_fields: list of tuples, optional
        A flat (non-nested) list of tuples indicating any extra fields present.
        Similar to a record's layout field; but cannot be nested.
    """

    def __init__(self, payload_width=8, valid_width=1, extra_fields=None):
        """
        Parameter:
            payload_width -- The width of the payload packets.
        """

        # If we don't have any extra fields, use an empty list in its place.
        if extra_fields is None:
            extra_fields = []

        # ... store our extra fields...
        self._extra_fields = extra_fields

        # ... and create our basic stream.
        super().__init__([
            ('valid',    valid_width),
            ('ready',    1),

            ('first',    1),
            ('last',     1),

            ('payload',  payload_width),
            *extra_fields
        ])


    def attach(self, interface, omit=None):
        # Create lists of fields to be copied -to- the interface (RHS fields),
        # and lists of fields to be copied -from- the interface (LHS fields).
        rhs_fields = ['valid', 'first', 'last', 'payload', *self._extra_fields]
        lhs_fields = ['ready']
        assignments = []

        if omit:
            rhs_fields = [field for field in rhs_fields if field not in omit]
            lhs_fields = [field for field in lhs_fields if field not in omit]


        # Create each of our assignments.
        for field in rhs_fields:
            assignment = interface[field].eq(self[field])
            assignments.append(assignment)
        for field in lhs_fields:
            assignment = self[field].eq(interface[field])
            assignments.append(assignment)

        return assignments

    def connect(self, interface, omit=None):
        return self.attach(interface, omit=omit)


    def stream_eq(self, interface, *, omit=None):
        """ A hopefully more clear version of .connect() that more clearly indicates data_flow direction.

        This will either solve a common footgun or introduce a new one. We'll see and adapt accordingly.
        """
        return interface.attach(self, omit=omit)


    def tap(self, interface, *, tap_ready=False, **kwargs):
        """ Simple extension to stream_eq() that captures a read-only view of the stream.

        This connects all signals from ``interface`` to their equivalents in this stream.
        """
        core = self.stream_eq(interface, omit={"ready"}, **kwargs)

        if tap_ready:
            core.append(self.ready.eq(interface.ready))

        return core




    def __getattr__(self, name):

        # Allow "data" to be a semantic alias for payload.
        # In some cases, this makes more sense to write; so we'll allow either.
        # Individual sections of the code base should stick to one or the other (please).
        if name == 'data':
            name = "payload"

        return super().__getattr__(name)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Low-level USB3 transciever gateware -- control transfer components. """

from amaranth import *

from usb_protocol.emitters    import DeviceDescriptorCollection
from usb_protocol.types       import USBRequestType, USBDirection

from ..protocol.endpoint      import SuperSpeedEndpointInterface
from ..application.request    import SuperSpeedRequestHandlerInterface, SuperSpeedSetupDecoder
from ..application.request    import SuperSpeedRequestHandlerMultiplexer, StallOnlyRequestHandler
from ..request.standard       import StandardRequestHandler


class USB3ControlEndpoint(Elaboratable):
    """ Gateware that manages control request data progression.

    This class is used by creating one or more *request handler* modules; which define how requests
    are handled. These handlers can be bound using :attr:`add_request_handler`.

    For convenience, this module can also automatically be populated with a ``StandardRequestHandler``
    via the :attr:`add_standard_request_handlers`.

    Attributes
    ----------
    interface: SuperSpeedEndpointInterface
        The interface from this endpoint to the core device hardware.

    Parameters
    ----------
        endpoint_number: int, optional
            The endpoint number for this control interface; defaults to (and almost always should
            be) zero.
    """

    def __init__(self, *, endpoint_number=0):
        self._endpoint_number = endpoint_number

        # List of the modules that will handle control requests.
        self._request_handlers = []

        #
        # I/O Port
        #
        self.interface = SuperSpeedEndpointInterface()



    def add_request_handler(self, request_handler):
        """ Adds a ControlRequestHandler module to this control endpoint.

        No arbitration is performed between request handlers; so it's important
        that request handlers not overlap in the requests they handle.
        """
        self._request_handlers.append(request_handler)


    def add_standard_request_handlers(self, descriptors: DeviceDescriptorCollection):
        """ Adds a handlers for the standard USB requests.

        This will handle all Standard-type requests; so any additional request handlers
        must not handle Standard requests.

        Parameters:

        """
        handler = StandardRequestHandler(descriptors)
        self._request_handlers.append(handler)


    def elaborate(self, platform):
        m = Module()

        # Shortcuts.
        interface      = self.interface
        handshakes_out = self.interface.handshakes_out
        handshakes_in  = self.interface.handshakes_in


        #
        # Convenience feature:
        #
        # If we have -only- a standard request handler, automatically add a handler that will
        # stall all other requests.
        #
        single_handler = (len(self._request_handlers) == 1)
        if (single_handler and isinstance(self._request_handlers[0], StandardRequestHandler)):

            # Add a handler that will stall any non-standard request.
            stall_condition = lambda setup : setup.type != USBRequestType.STANDARD
            self.add_request_handler(StallOnlyRequestHandler(stall_condition))




        #
        # Setup packet decoder.
        #
        m.submodules.setup_decoder = setup_decoder = SuperSpeedSetupDecoder()
        m.d.comb += [
            setup_decoder.sink       .tap(interface.rx),
            setup_decoder.header_in  .eq(interface.rx_header),

            setup_decoder.rx_good    .eq(interface.rx_complete),
            setup_decoder.rx_bad     .eq(interface.rx_invalid),
        ]


        #
        # Request handler interfacing.
        #

        # Multiplex the output of each of our request handlers.
        m.submodules.request_mux = request_mux = SuperSpeedRequestHandlerMultiplexer()
        request_interface = request_mux.shared

        # Add each of our handlers to the endpoint; and add it to our mux.
        for handler in self._request_handlers:

            # Create a display name for the handler...
            name = handler.__class__.__name__
            if hasattr(m.submodules, name):
                name = f"{name}_{id(handler)}"

            # ... and add it.
            m.submodules[name] = handler
            request_mux.add_interface(handler.interface)


        # To simplify the request-handler interface, we'll only pass through our Rx stream
        # when the most recently header packet targets our endpoint number.
        with m.If(interface.rx_header.endpoint_number == self._endpoint_number):
            m.d.comb += [
                request_interface.rx           .tap(interface.rx),
                request_interface.rx_complete  .eq(interface.rx_complete),
                request_interface.rx_invalid   .eq(interface.rx_invalid),
            ]

        # The remainder of our signals are always hooked up.
        m.d.comb += [
            request_interface.rx_header    .eq(interface.rx_header),

            # Transmit. Note that our transmit direction is always set to OUT; even though we're
            # sending data to the host, per [USB3.2r1: 8.12.2].
            interface.tx                   .stream_eq(request_interface.tx),
            interface.tx_length            .eq(request_interface.tx_length),
            interface.tx_sequence_number   .eq(0),
            interface.tx_endpoint_number   .eq(self._endpoint_number),
            interface.tx_direction         .eq(USBDirection.OUT),

            # Status.
            interface.handshakes_in        .connect(request_interface.handshakes_in),
            interface.handshakes_out       .connect(request_interface.handshakes_out),

            # Address / config management.
            interface.address_changed      .eq(request_interface.address_changed),
            interface.new_address          .eq(request_interface.new_address),

            interface.config_changed       .eq(request_interface.config_changed),
            interface.new_config           .eq(request_interface.new_config),

        ]

        #
        # Data sequence handling.
        #

        #
        # SETUP stage handling.
        #

        # Our setup stage is easy: we just pass through our setup packet through to the request handlers...
        m.d.comb += request_interface.setup.eq(setup_decoder.packet)

        # ... and then ACK any SETUP packet that comes through, as we're required to [USB3.2r1: 8.12.2].
        # Note that our SETUP sequence number needs to be `1`.
        with m.If(setup_decoder.packet.received):
            m.d.comb += [
                handshakes_out.retry_required   .eq(0),
                handshakes_out.next_sequence    .eq(1),
                handshakes_out.send_ack         .eq(1),
            ]


        # Always set the endpoint number in our handshakes.
        m.d.comb += handshakes_out.endpoint_number  .eq(self._endpoint_number),

        #
        # DATA stage handling.
        #
        is_to_us    = (handshakes_in.endpoint_number == self._endpoint_number)
        is_in_token = (handshakes_in.number_of_packets != 0)

        # Pass through any IN tokens as data requests to our request handlers.
        with m.If(handshakes_in.ack_received & is_to_us & is_in_token):
            m.d.comb += request_interface.data_requested.eq(1)


        #
        # STATUS stage handling.
        #
        with m.If(handshakes_in.status_received):
            m.d.comb += request_interface.status_requested.eq(1)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Full-device test harnesses for USB2. """

from usb_protocol.types import USBStandardRequests
from ..usb.usb2         import USBPacketID

from .                  import LunaGatewareTestCase

from .contrib           import usb_packet
from ..interface.utmi   import UTMIInterface

class USBDeviceTest(LunaGatewareTestCase):
    """ Test case strap for UTMI-connected devices. """

    # Use only the USB clock domain; ignore the sync one unless overridden.
    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY  = 60e6

    # The name of the argument to the DUT that will accept our UTMI bus.
    UTMI_BUS_ARGUMENT = 'bus'

    # The maximum number of NAKs we should allow in a loop before we bail out.
    # Keeping this reasonable prevents the simulation from running indefinitely.
    MAX_NAKS = 100

    def instantiate_dut(self):
        self.utmi    = UTMIInterface()

        # Vitals about the device.
        self.address             = 0
        self.max_packet_size_ep0 = 64

        # Always pass in the UTMI bus.
        arguments = self.FRAGMENT_ARGUMENTS.copy()
        arguments[self.UTMI_BUS_ARGUMENT] = self.utmi

        dut =  self.FRAGMENT_UNDER_TEST(**arguments)
        self.provision_dut(dut)

        return dut


    def provision_dut(self, dut):
        """ Hook that allows us to add any desired properties to the DUT before simulation.

        This method is called before initial elaboration; so functions that modify devices
        before elaboration can be used.
        """
        pass


    def provide_byte(self, byte):
        """ Provides a given byte on the UTMI receive data for one cycle. """
        yield self.utmi.rx_data.eq(byte)
        yield


    def start_packet(self, *, set_rx_valid=True):
        """ Starts a UTMI packet receive. """
        yield self.utmi.rx_active.eq(1)

        if set_rx_valid:
            yield self.utmi.rx_valid.eq(1)

        yield


    def end_packet(self):
        """ Starts a UTMI packet receive. """
        yield self.utmi.rx_active.eq(0)
        yield self.utmi.rx_valid.eq(0)
        yield


    def provide_packet(self, *octets, cycle_after=True):
        """ Provides an entire packet transaction at once. """

        yield from self.start_packet()

        for b in octets:
            yield from self.provide_byte(b)

        yield from self.end_packet()

        if cycle_after:
            yield


    @staticmethod
    def bits_to_octets(bits):
        """ Converts a string of bits to octets.

        For compatibility with antmicro's USB-testbench functions.
        """

        bits   = bits[:]
        octets = []

        assert (len(bits) % 8) == 0

        while bits:
            # Grab an octet worth of bits..
            octet_bits = bits[0:8][::-1]
            bits = bits[8:]

            # ... and parse it.
            octet = int(octet_bits, 2)
            octets.append(octet)

        return octets


    def provide_bits(self, bits, cycle_after=True):
        """ Provides an entire packet transaction at once; accepts bits. """

        octets = self.bits_to_octets(bits)
        yield from self.provide_packet(*octets, cycle_after=cycle_after)



    def send_token(self, pid, *, endpoint=0, address=None):
        """ Issues a token packet to the simulated USB device.

        Parameters:
            pid      -- The PID of the packet to be sent.
            endpoint -- The endpoint on which the token should be sent.
            address  -- The address of the device to be targeted.
                        If omitted, the most recently set-address'd address is used.
        """

        # Grab the raw bits that make up our token from the Antmicro library...
        address = address if address else self.address
        bits    = usb_packet.token_packet(pid, address, endpoint)

        # ... and issue them ourselves.
        yield from self.provide_bits(bits)



    def send_data(self, pid, *octets):
        """ Sends a data packet to the simulated USB device.

        Parameters:
            pid  -- The PID to send the provided packet with.
            *data -- The data to be sent.
        """

        bits = usb_packet.data_packet(pid, octets)
        yield from self.provide_bits(bits)




    def send_handshake(self, pid):
        """ Issues a handshake packet to the simulated USB device.

        Parameters:
            pid      -- The PID of the packet to be sent.
        """

        # Ensure we have an USBPacketID-wrapped PID.
        pid = USBPacketID(pid)
        yield from self.provide_packet(pid.byte())




    def receive_packet(self, as_bytes=True, timeout=1000):
        """ Receives a collection of data from the USB bus. """

        data = []

        # Mark us as ready to receive data.
        yield self.utmi.tx_ready.eq(1)
        yield

        # Wait for the TX line to be valid.
        yield from self.wait_until(self.utmi.tx_valid, timeout=timeout)

        # For as long as tx_valid is asserted, capture data.
        while (yield self.utmi.tx_valid):
            data.append((yield self.utmi.tx_data))
            yield

        # Mark us as ready to receive data.
        yield self.utmi.tx_ready.eq(0)

        return bytes(data) if as_bytes else data


    #
    # More complex transaction helpers.
    #

    def interpacket_delay(self):
        """ Waits for a period appropriate between each packet. """

        # We'll use a shorter value than real interpacket delays here,
        # to speed up simulation. This can be tuned longer if necessary.
        # A spec-valued length would be 10 for a FS device, or 1 for a HS.
        yield from self.advance_cycles(1)


    def out_transaction(self, *octets, endpoint=0, token_pid=USBPacketID.OUT,
        data_pid=USBPacketID.DATA0, expect_handshake=None):
        """ Performs an OUT transaction.

        Parameters:
            *octets          -- The data to send.
            endpoint         -- The endpoint on which to send the relevant data.
            token_pid        -- The token PID to send. Defaults to OUT.
            data_pid         -- The data PID to send. Defalts to DATA0.
            expect_handshake -- If provided, we'll assert that a given handshake is provided.

        Returns the handshake received.
        """

        # Issue the token...
        yield from self.send_token(token_pid, endpoint=endpoint)
        yield from self.interpacket_delay()

        # ... issue our DATA packet...
        yield from self.send_data(data_pid, *octets)

        # ... and receive our handshake.
        data = yield from self.receive_packet()

        # If we expect the SETUP packet to be ACK'd, validate that.
        if expect_handshake:
            self.assertEqual(USBPacketID.from_byte(data), expect_handshake)

        yield from self.interpacket_delay()
        return USBPacketID.from_byte(data)



    def out_transfer(self, *octets, endpoint=0, data_pid=USBPacketID.DATA0, max_packet_size=64):
        """ Performs an OUT transaction.

        Parameters:
            *octets          -- The data to send.
            endpoint         -- The endpoint on which to send the relevant data.
            data_pid         -- The first data PID to send. Defalts to DATA0.
            max_packet_size  -- The maximum packet size for the current endpoint.

        Returns the final handshake received.
        """

        to_send  = octets[:]

        # If we're going to send all max-packet-length sized packet,
        # we'll add a ZLP.
        send_zlp = (len(octets) % max_packet_size) == 0

        while to_send:

            # Pull a packet out of the send queue...
            packet = to_send[0:max_packet_size]

            # ... and try to send it.
            handshake = yield from self.out_transaction(endpoint=endpoint, data_pid=data_pid, *packet)

            # If we're stalled, abort immediately.
            if handshake == USBPacketID.STALL:
                return USBPacketID.STALL

            # If our packet is NAK'd, don't continue.
            if handshake == USBPacketID.NAK:
                continue

            # Otherwise, advance in the stream...
            to_send = to_send[max_packet_size:]

            # ... and toggle DATA PIDs.
            data_pid = USBPacketID.DATA1 if (data_pid == USBPacketID.DATA0) else USBPacketID.DATA0


        # If we're going to send a ZLP, send it.
        if send_zlp:
            handshake = USBPacketID.NAK
            while handshake == USBPacketID.NAK:
                handshake = yield from self.out_transaction(endpoint=endpoint, data_pid=data_pid)

        return handshake



    def in_transaction(self, endpoint=0, data_pid=None, handshake=USBPacketID.ACK):
        """ Performs an IN transaction.

        Parameters:
            endpoint    -- The endpoint on which to fetch the relevant data.
            data_pid    -- The data PID to expect, or None if we don't are.
            handshake   -- The response we should give after receiving data.

        Returns:
            handshake   -- The handshake retrieved in response.
            data        -- A list of octets received.
        """

        # Issue the IN token...
        yield from self.send_token(USBPacketID.IN, endpoint=endpoint)

        # ... receive our response...
        result = yield from self.receive_packet()
        try:
            pid, *data, crc_low, crc_high = result

        # If we can't unpack the response, just return the PID.
        except ValueError:
            return USBPacketID.from_int(result[0]), None

        # ... validate its CRC...
        expected_crc = usb_packet.crc16(data)
        self.assertEqual([crc_low, crc_high], expected_crc)

        # ... validate pid toggling, if desired...
        if data_pid:
            self.assertEqual(pid, data_pid.byte())

        # ... and issue our handshake.
        yield from self.interpacket_delay()
        yield from self.send_handshake(handshake)

        yield from self.interpacket_delay()
        return USBPacketID.from_int(pid), data


    def in_transfer(self, endpoint=0, data_pid=None, handshake=USBPacketID.ACK):
        """ Performs an IN transaction.

        Parameters:
            endpoint    -- The endpoint on which to fetch the relevant data.
            data_pid    -- The data PID to expect, or None if we don't are.
            handshake   -- The response we should give after receiving data.

        Returns:
            handshake   -- The last data handshake received.
            data        -- A list of octets received.
        """

        naks = 0
        data = []

        while True:
            pid, packet = yield from self.in_transaction(
                endpoint=0, data_pid=data_pid, handshake=USBPacketID.ACK)

            # If we were NAK'd, try again.
            if pid == USBPacketID.NAK:
                naks += 1
                self.assertLess(naks, self.MAX_NAKS)

                continue

            # If we were stalled, abort here.
            if pid == USBPacketID.STALL:
                return pid, None

            # Add the given packet to our transaction.
            data.extend(packet)

            # Swap to expecting the next packet ID.
            data_pid = USBPacketID.DATA1 if (data_pid == USBPacketID.DATA0) else USBPacketID.DATA0

            # If this is a short packet, stop receiving.
            if len(packet) < self.max_packet_size_ep0:
                break

        return pid, data


    def setup_transaction(self, request_type, request, value=0, index=0, length=0):
        """ Sends a SETUP transaction. All arguments match their SETUP packet definitions.

        Returns the handshake received.
        """

        def split(arg):
            """ Convenience function that splits a setup parameter. """
            high = arg >> 8
            low  = arg & 0xFF

            return low, high

        # ... issue our DATA token...
        response = yield from self.out_transaction(
            request_type, request, *split(value), *split(index), *split(length),
            token_pid=USBPacketID.SETUP, expect_handshake=USBPacketID.ACK
        )

        return response


    def control_interphase_delay(self):
        """ Waits for a period appropriate between each delays. """

        # This is shorter than would be normal, in order to speed up simulation.
        # If necessary, this can be extended arbitrarily.
        yield from self.advance_cycles(1)


    def control_request_in(self, request_type, request, value=0, index=0, length=0):
        """ Performs an IN control request, and returns the results.

        Arguments match the SETUP packets.

        Returns:
            handshake -- The handshake value returned.
            data      -- A list of octets returned.
        """

        naks = 0

        # If we don't have a data phase, treat this identically to an OUT control request.
        if length == 0:
            response = yield from self.control_request_out(request_type, request,
                value=value, index=index)
            return response

        #
        # Issue the Setup phase.
        #
        yield from self.setup_transaction(request_type, request, value, index, length)
        yield from self.control_interphase_delay()

        #
        # Issue the Data phase.
        #
        pid, packet = yield from self.in_transfer(data_pid=USBPacketID.DATA1)
        if pid == USBPacketID.STALL:
            return pid, None

        #
        # Issue the Status phase.
        #
        yield from self.control_interphase_delay()

        # Read until we get a status other than a NAK.
        handshake = USBPacketID.NAK
        while handshake == USBPacketID.NAK:
            naks += 1
            self.assertLess(naks, self.MAX_NAKS)

            handshake = yield from self.out_transaction(data_pid=USBPacketID.DATA1)


        # Finally, return our handshake and our data.
        return handshake, packet


    def control_request_out(self, request_type, request, value=0, index=0, data=()):
        """ Performs an OUT control request, and returns the results.

        Arguments match the SETUP packets.
        """

        naks = 0

        #
        # Issue the Setup phase.
        #
        yield from self.setup_transaction(request_type, request, value, index, len(data))
        yield from self.control_interphase_delay()


        #
        # Issue the Data phase, if we have one.
        #
        if data:
            pid = yield from self.out_transfer(data_pid=USBPacketID.DATA1, *data)
            if pid == USBPacketID.STALL:
                return pid


        #
        # Issue the Status phase.
        #
        yield from self.control_interphase_delay()

        # We'll perform an IN transfer, and expect a ZLP back.
        pid = USBPacketID.NAK
        while pid == USBPacketID.NAK:
            naks += 1
            self.assertLess(naks, self.MAX_NAKS)

            pid, data = yield from self.in_transaction(data_pid=USBPacketID.DATA1)

        # If this wasn't stalled, ensure we get a ZLP back.
        if pid != USBPacketID.STALL:
            self.assertEqual(data, [])

        # Finally, return our handshake.
        return pid


    def get_descriptor(self, descriptor_type, index=0, length=64):
        """ Performs a GET_DESCRIPTOR request; fetching a USB descriptor.

        Parameters:
            descriptor_type -- The descriptor type number to fetch.
            index           -- The descriptor index to fetch.
            length          -- The requested length to read.
        """

        # The type is stored in the MSB of the value; and the index in the LSB.
        value = descriptor_type << 8 | index
        descriptor = yield from self.control_request_in(0x80,
                USBStandardRequests.GET_DESCRIPTOR, value=value, length=length)

        return descriptor


    def set_address(self, new_address, update_address=True):
        """ Performs a SET_ADDRESS request; setting the device address.

        Parameters:
            new_address -- The address to apply.
        """

        response_pid = yield from self.control_request_out(0,
            USBStandardRequests.SET_ADDRESS, value=new_address)

        if update_address and (response_pid == USBPacketID.DATA1):
            self.address = new_address

        return response_pid


    def set_configuration(self, number):
        """ Performs a SET_CONFIGURATION request; configuring the device.

        Parameters:
            number -- The configuration to select.
        """
        response_pid = yield from self.control_request_out(0,
            USBStandardRequests.SET_CONFIGURATION, value=number)
        return response_pid


    def get_configuration(self):
        """ Performs a GET_CONFIGURATION request; reading the device's configuration. """
        response = yield from self.control_request_in(0x80,
            USBStandardRequests.GET_CONFIGURATION, length=1)
        return response



#
# This file is part of LUNA.
#
# Copyright (c) 2025 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Core stream definitions for supporting native Amaranth 0.5 streams. """

from amaranth      import *
from amaranth.lib  import data

class Packet(data.StructLayout):
    def __init__(self, data_layout, first=True, last=True):
        layout = (first and { "first": unsigned(1) } or {}) \
               | (last  and { "last":  unsigned(1) } or {})
        super().__init__(layout | {
            "data": data_layout
        })



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code adapted from ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Code for handling SKP ordered sets on the transmit and receive path.

SKP ordered sets are provided in order to give some "padding data" that can be removed
in order to handle differences in transmitter/receiver clock rates -- a process called
"clock tolerance compensation" (CTC). The actual insertion and removal of SKP ordered sets
for CTC is handled by the PHY -- but it only adds and removes sets where it needs to to
compensate for clock differences.

It's up to us to insert and remove additional ordered sets.
"""

from amaranth import *

from .coding import SKP, stream_word_matches_symbol
from ...stream import USBRawSuperSpeedStream

class CTCSkipRemover(Elaboratable):
    """ Clock Tolerance Compensation (CTC) receive buffer gateware.

    It's functionally impossible to precisely synchronize the clocks for two independent
    systems -- every specification has to allow for some difference in frequency between
    the system's clocks (the "clock tolerance")

    To compensate, high speed serial protocols inject 'filler' bytes called "SKP ordered sets",
    which can be safely discarded. This allows the slower clock to catch up to the faster one.
    [USB 3.2r1: 6.4.3].

    Our PHY handles the core clock tolerance compesnation inside of its own clock domain; removing
    these filler sets whenever removing them helps to keep the receiver and transmitter's clocks in sync.
    This leaves behind the sets whose removal would not directly help with CTC.

    This module removes those leftovers before data leaves the physical layer.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input stream
        The stream from which SKP ordered sets should be removed.
    source: USBRawSuperSpeedStream(), output stream
        The relevant stream with SKP ordered sets removed. Note that past this point,
        ``stream.valid`` can and will sometimes be false.

    skip_removed: Signal(), output
        Strobe that indicates that a SKP ordered set was removed.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.sink            = USBRawSuperSpeedStream()
        self.source          = USBRawSuperSpeedStream()

        self.skip_removed    = Signal()
        self.bytes_in_buffer = Signal(range(len(self.sink.ctrl) + 1))


    def elaborate(self, platform):
        m = Module()

        sink   = self.sink
        source = self.source

        bytes_in_stream = len(sink.ctrl)

        #
        # Find SKP symbols
        #

        # Identify the locations of any SKP symbols present in the stream.
        skp_locations = Signal(bytes_in_stream)
        for i in range(bytes_in_stream):
            m.d.comb += skp_locations[i].eq(stream_word_matches_symbol(sink, i, symbol=SKP))

        # If we've found one, indicate that we're removing it.
        skip_found = self.sink.valid & self.sink.ready & (skp_locations != 0)
        m.d.comb += self.skip_removed.eq(skip_found)


        #
        # Data Extractor
        #

        # We'll first extract the data and control bits for every position that doesn't contain a SKP.
        valid_data        = Signal.like(sink.data)
        valid_ctrl        = Signal.like(sink.ctrl)
        valid_byte_count  = Signal(range(0, bytes_in_stream + 1))

        # We have a SKIP location for each byte; and each locations has two possible values
        # (SKP, no SKP); and therefore we have 2 ** <bytes> distinct arrangements.
        possible_arrangement_count = 2 ** bytes_in_stream
        possible_arrangements = range(possible_arrangement_count)

        # We'll handle each possibility with a programmatically generated case.
        with m.Switch(skp_locations):

            # We'll generate a single case for each possible "skip mask".
            for skip_mask in possible_arrangements:
                with m.Case(skip_mask):
                    data_fragments = []
                    ctrl_fragments = []

                    # We'll iterate over each of our possible positions, and gather
                    # the Amaranth signals associated with the non-skip values in the
                    # relevant position.
                    for position in range(bytes_in_stream):

                        # If this case would have a valid byte at the given position, grab it.
                        position_mask = 1 << position
                        if (position_mask & skip_mask) == 0:
                            data_signal_at_position = sink.data.word_select(position, 8)
                            ctrl_signal_at_position = sink.ctrl[position]
                            data_fragments.append(data_signal_at_position)
                            ctrl_fragments.append(ctrl_signal_at_position)


                    # If there are any valid data signals associated with the given position,
                    # coalesce the data and control signals into a single word, which we'll handle below.
                    if data_fragments:
                        m.d.comb += [
                            valid_data.eq(Cat(*data_fragments)),
                            valid_ctrl.eq(Cat(*ctrl_fragments)),
                            valid_byte_count.eq(len(data_fragments)),
                        ]

        #
        # Elastic Buffer / Valid Data Coalescence
        #

        # We now have a signal that contains up to a valid word of data. We'll need to
        # stitch this data together before we can use it. To do so, we'll use a shift
        # register long enough to store two complete words of data -- one for the word
        # we're outputting, and one for a word-in-progress.

        # This shift register serves as a local "elastic buffer" -- we can add in data in
        # bits and pieces, and remove it in bits and pieces.
        buffer_size_bytes = bytes_in_stream * 2

        # Create our internal shift register, as well as our current fill counter.
        data_buffer     = Signal(buffer_size_bytes * 8)
        ctrl_buffer     = Signal(buffer_size_bytes)
        bytes_in_buffer = Signal(range(0, buffer_size_bytes + 1))

        # Determine if we'll have a valid stream
        m.d.comb += sink.ready.eq(bytes_in_buffer <= buffer_size_bytes)

        # If we're receiving data this round, add it into our shift register.
        with m.If(sink.valid & sink.ready):

            # Compute how many bytes we'll have next cycle: it's the number of bytes we already have
            # (bytes_in_buffer) plus the bytes we're adding (valid_byte_count) and minus the data we're
            # about to remove (one word, or bytes_in_stream) if we're reading, or minus nothing if we're not.
            with m.If(source.valid & source.ready):
                m.d.ss += bytes_in_buffer.eq(bytes_in_buffer + valid_byte_count - bytes_in_stream)
            with m.Else():
                m.d.ss += bytes_in_buffer.eq(bytes_in_buffer + valid_byte_count)

            # Handle our shift register pushing logic.
            with m.Switch(valid_byte_count):

                # Our simplest case: we have no data in the buffer; and nothing needs to change.
                with m.Case(0):
                    pass

                # In every other case, we have some data to be added to the buffer.
                # We'll do the math slightly differently for each potential number of bytes.
                for i in range(1, bytes_in_stream + 1):
                    with m.Case(i):

                        # Grab our existing data, and stick it onto the end of the shift register.
                        m.d.ss += [
                            data_buffer  .eq(Cat(data_buffer[8*i:], valid_data[0:8*i])),
                            ctrl_buffer  .eq(Cat(ctrl_buffer[1*i:], valid_ctrl[0:1*i])),
                        ]

        # If we're not receiving data, but we -are- removing it, we'll just update our total
        # valid data counter to account for the removal.
        with m.Elif(source.valid & source.ready):
            m.d.ss += bytes_in_buffer.eq(bytes_in_buffer - bytes_in_stream)


        #
        # Data output
        #

        # We'll output a word each time we have enough data in our shift register toS
        # output a full word.
        m.d.comb += source.valid.eq(bytes_in_buffer >= bytes_in_stream)

        # Our data ends in different places depending on how many bytes we
        # have in our shift register; so we'll need to pop it from different locations.
        with m.Switch(bytes_in_buffer):
            for i in range(bytes_in_stream, bytes_in_stream * 2):
                with m.Case(i):
                    # Grab the relevant word from the end of the buffer.
                    word_position = 8 - i
                    m.d.comb += [
                        source.data.eq(data_buffer[8 * word_position : 8 * (word_position + bytes_in_stream)]),
                        source.ctrl.eq(ctrl_buffer[1 * word_position : 1 * (word_position + bytes_in_stream)]),
                    ]

        #
        # Diagnostic output.
        #
        m.d.comb += self.bytes_in_buffer.eq(bytes_in_buffer)

        return m


class CTCSkipInserter(Elaboratable):
    """ Clock Tolerance Compensation (CTC) Skip insertion gateware.

    See the ``CTCSkipRemover`` for a description of CTC and its general operation.

    Our PHY handles the core clock tolerance compesnation inside of its own clock domain; adding
    Skip sets whenever adding them helps to keep the transmitter's elastic buffer from running low
    on data. However, we still need to add in our own Skip ordered sets so the other side of the link
    has enough to perform its own CTC adjustments.

    This module adds ordered sets, per the USB standard.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input stream
        The stream into which SKP ordered sets should be inserted.
    source: USBRawSuperSpeedStream(), output stream
        The relevant stream with SKP ordered sets inserted.

    can_send_skip: Signal(), input
        Controls when SKPs can be inserted. This should be asserted when we're transmitting
        logical idle.

    sending_skip: Signal(), output
        Indicates that we're currently sending only SKP characters; and thus our scrambler
        should not advance.
    """

    SKIP_BYTE_LIMIT = 354

    def __init__(self):
        #
        # I/O port
        #
        self.sink          = USBRawSuperSpeedStream()
        self.source        = USBRawSuperSpeedStream()

        self.can_send_skip = Signal()
        self.sending_skip  = Signal()


    def elaborate(self, platform):
        m = Module()

        sink   = self.sink
        source = self.source

        #
        # SKP scheduling.
        #

        # The largest amount of pending SKP ordered sets can right before finishing transmitting:
        #   (20 bytes of DPH) + (1036 bytes of DPP) + (6 bytes of SKP)
        # This sequence is 1062, or 354*3, bytes long. Since we only transmit pairs of SKP ordered sets,
        # the maximum amount of pending SKP ordered sets at any time is 4.
        skips_to_send = Signal(range(5))
        skip_needed   = Signal()

        # Precisely count the amount of skip ordered sets that will be inserted at the next opportunity.
        # From [USB3.0r1: 6.4.3]: "The non-integer remainder of the Y/354 SKP calculation shall not be
        # discarded and shall be used in the calculation to schedule the next SKP Ordered Set."
        with m.If(skip_needed & ~self.sending_skip):
            m.d.ss += skips_to_send.eq(skips_to_send + 1)
        with m.If(~skip_needed & self.sending_skip):
            m.d.ss += skips_to_send.eq(skips_to_send - 2)
        with m.If(skip_needed & self.sending_skip):
            m.d.ss += skips_to_send.eq(skips_to_send - 1)


        #
        # SKP insertion timing.
        #
        bytes_per_word = len(self.sink.ctrl)
        data_bytes_elapsed = Signal(range(self.SKIP_BYTE_LIMIT))

        # Count each byte of data we send...
        with m.If(sink.valid & sink.ready):
            m.d.ss += data_bytes_elapsed.eq(data_bytes_elapsed + bytes_per_word)

            # ... and once we see enough data, schedule insertion of a skip ordered set.
            with m.If(data_bytes_elapsed + bytes_per_word >= self.SKIP_BYTE_LIMIT):
                m.d.ss   += data_bytes_elapsed.eq(data_bytes_elapsed + bytes_per_word - self.SKIP_BYTE_LIMIT)
                m.d.comb += skip_needed.eq(1)


        #
        # SKP insertion.
        #

        # Finally, if we can send a skip this cycle and need to, replace our IDLE with two SKP ordered sets.
        #
        # Although [USB3.0r1: 6.4.3] allows "during training only [...] the option of waiting to insert 2 SKP
        # ordered sets when the integer result of Y/354 reaches 2", inserting individual SKP ordered sets on
        # a 32-bit data path has considerable overhead, and we only insert pairs.
        with m.If(self.can_send_skip & (skips_to_send >= 2)):
            m.d.comb += self.sending_skip.eq(1)
            m.d.ss += [
                source.valid       .eq(1),
                source.data        .eq(SKP.value_const().replicate(len(source.ctrl))),
                source.ctrl        .eq(SKP.ctrl_const().replicate(len(source.ctrl))),
            ]

        with m.Else():
            m.d.ss += [
                self.source        .stream_eq(self.sink),
            ]

        return m



# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
#
# Based on I2C code from Glasgow
# I2C reference: https://www.nxp.com/docs/en/user-guide/UM10204.pdf

from amaranth import Elaboratable, Module, Signal, Cat, C
from amaranth.lib.cdc import FFSynchronizer
from amaranth.hdl.rec import Record, DIR_FANIN, DIR_FANOUT


__all__ = ["I2CBus", "I2CInitiator", "I2CRegisterInterface"]

class I2CBus(Record):
    """ Record representing an I2C bus. """
    def __init__(self):
        super().__init__([
            ('scl', [('i', 1, DIR_FANIN), ('o', 1, DIR_FANOUT), ('oe', 1, DIR_FANOUT)]),
            ('sda', [('i', 1, DIR_FANIN), ('o', 1, DIR_FANOUT), ('oe', 1, DIR_FANOUT)]),
        ])

class I2CRegisterInterface(Elaboratable):
    """ Gateware interface that handles I2C register reads and writes.

    I/O ports:

        # Controller signals:
        O: busy              -- indicates when the interface is busy processing a transaction
        I: address[8]        -- the address of the register to work with
        O: done              -- strobe that indicates when a register request is complete

        I: read_request      -- strobe that requests a register read
        O: read_data[8]      -- data read from the relevant register read

        I: write_request     -- strobe that indicates a register write
        I: write_data[8]     -- data to be written during a register write

    """
    def __init__(self, pads, *, period_cyc, address, clk_stretch=False, data_bytes=1):

        self.pads          = pads
        self.period_cyc    = period_cyc
        self.dev_address   = address
        self.clk_stretch   = clk_stretch

        # I/O ports

        self.busy          = Signal()
        self.address       = Signal(8)
        self.size          = Signal(range(data_bytes+1))
        self.done          = Signal()

        self.read_request  = Signal()
        self.read_data     = Signal(8 * data_bytes)

        self.write_request = Signal()
        self.write_data    = Signal(8 * data_bytes)

    def elaborate(self, platform):
        m = Module()

        current_address = Signal.like(self.address, reset_less=True)
        current_write   = Signal.like(self.write_data, reset_less=True)
        current_read    = Signal.like(self.read_data - 8, reset_less=True)
        rem_bytes       = Signal.like(self.size, reset_less=True)
        is_write        = Signal(reset_less=True)

        # I2C initiator (low level manager) and default signal values
        m.submodules.i2c = i2c = I2CInitiator(pads=self.pads, period_cyc=self.period_cyc, clk_stretch=self.clk_stretch)
        m.d.comb += [
            i2c.start .eq(0),
            i2c.write .eq(0),
            i2c.read  .eq(0),
            i2c.stop  .eq(0),
        ]

        with m.FSM() as fsm:

            # We're busy whenever we're not IDLE; indicate so.
            m.d.comb += self.busy.eq(~fsm.ongoing('IDLE'))

            # IDLE: wait for a request to be made
            with m.State('IDLE'):
                with m.If(self.read_request | self.write_request):
                    m.d.sync += [
                        current_address .eq(self.address),
                        current_write   .eq(self.write_data),
                        current_read    .eq(0),
                        rem_bytes       .eq(self.size),
                        is_write        .eq(self.write_request),
                    ]
                    m.next = 'START'

            # Common device and register address handling
            with m.State('START'):
                with m.If(~i2c.busy):
                    m.d.comb += i2c.start.eq(1)
                    m.next = 'SEND_DEV_ADDRESS'

            with m.State("SEND_DEV_ADDRESS"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.data_i.eq((self.dev_address << 1) | 0),
                        i2c.write .eq(1),
                    ]
                    m.next = "ACK_DEV_ADDRESS"

            with m.State("ACK_DEV_ADDRESS"):
                with m.If(~i2c.busy):
                    with m.If(i2c.ack_o):  # dev address asserted
                        m.next = "SEND_REG_ADDRESS"
                    with m.Else():
                        m.next = "ABORT"

            with m.State("SEND_REG_ADDRESS"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.data_i.eq(current_address),
                        i2c.write .eq(1),
                    ]
                    m.next = "ACK_REG_ADDRESS"

            with m.State("ACK_REG_ADDRESS"):
                with m.If(~i2c.busy):
                    with m.If(~i2c.ack_o):
                        m.next = "ABORT"   # register address not asserted
                    with m.Elif(rem_bytes == 0):
                        m.next = "FINISH"  # 0-byte read/write
                    with m.Elif(is_write):
                        m.next = "WR_SEND_VALUE"
                    with m.Else():
                        m.next = "RD_START"

            # Write states
            # These handle the transmission of the successive bytes in the 
            # current write request

            with m.State("WR_SEND_VALUE"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.data_i.eq(current_write[-8:]),
                        i2c.write .eq(1),
                    ]
                    # prepare next byte too
                    m.d.sync += [
                        current_write.eq(current_write << 8),
                        rem_bytes    .eq(rem_bytes - 1),
                    ]
                    m.next = "WR_ACK_VALUE"
            
            with m.State("WR_ACK_VALUE"):
                with m.If(~i2c.busy):
                    with m.If(~i2c.ack_o):
                        m.next = "ABORT"
                    with m.Elif(rem_bytes == 0):
                        m.next = "FINISH"
                    with m.Else():
                        m.next = "WR_SEND_VALUE"

            # Read states
            # Once the source register address is written in the common states,
            # the following handles the retrieval of bytes from the device with
            # a new I2C read request.

            with m.State('RD_START'):
                with m.If(~i2c.busy):
                    m.d.comb += i2c.start.eq(1)
                    m.next = 'RD_SEND_DEV_ADDRESS'

            with m.State("RD_SEND_DEV_ADDRESS"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.data_i.eq((self.dev_address << 1) | 1),
                        i2c.write .eq(1),
                    ]
                    m.next = "RD_ACK_DEV_ADDRESS"

            with m.State("RD_ACK_DEV_ADDRESS"):
                with m.If(~i2c.busy):
                    with m.If(i2c.ack_o):  # dev address asserted
                        m.next = "RD_RECV_VALUE"
                    with m.Else():
                        m.next = "ABORT"

            with m.State("RD_RECV_VALUE"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.ack_i.eq(~(rem_bytes == 1)),  # 0 in last read byte
                        i2c.read .eq(1),
                    ]
                    m.d.sync += rem_bytes.eq(rem_bytes - 1)
                    m.next = "RD_WAIT_VALUE"

            with m.State("RD_WAIT_VALUE"):
                with m.If(~i2c.busy):
                    m.d.sync += current_read.eq((current_read << 8) | i2c.data_o)
                    with m.If(rem_bytes == 0):
                        m.d.sync += self.read_data.eq((current_read << 8) | i2c.data_o)
                        m.next = "FINISH"
                    with m.Else():
                        m.next = "RD_RECV_VALUE"

            # Common "exit" states that return to idle
            # FINISH asserts "done" to tell the transfer was done successfully
            with m.State("FINISH"):
                with m.If(~i2c.busy):
                    m.d.comb += [
                        i2c.stop .eq(1),
                        self.done.eq(1),
                    ]
                    m.next = "IDLE"
            
            with m.State("ABORT"):
                with m.If(~i2c.busy):
                    m.d.comb += i2c.stop .eq(1)
                    m.next = "IDLE"

        return m


class I2CBusDriver(Elaboratable):
    """
    Decodes bus conditions (start, stop, sample and setup) and provides synchronization.
    """
    def __init__(self, pads):
        self.scl_t  = pads.scl_t if hasattr(pads, "scl_t") else pads.scl
        self.sda_t  = pads.sda_t if hasattr(pads, "sda_t") else pads.sda

        self.scl_i  = Signal()
        self.scl_o  = Signal(init=1)
        self.sda_i  = Signal()
        self.sda_o  = Signal(init=1)

        self.sample = Signal(name="bus_sample")
        self.setup  = Signal(name="bus_setup")
        self.start  = Signal(name="bus_start")
        self.stop   = Signal(name="bus_stop")

    def elaborate(self, platform):
        m = Module()

        # SDA line must be bidirectional...
        m.d.comb += [
            self.sda_t.o  .eq(0),
            self.sda_t.oe .eq(~self.sda_o),
        ]
        m.submodules += FFSynchronizer(self.sda_t.i, self.sda_i, init=1)

        # But the SCL line does not need to: only if we want to support clock stretching
        if hasattr(self.scl_t, "oe"):
            m.d.comb += [
                self.scl_t.o  .eq(0),
                self.scl_t.oe .eq(~self.scl_o),
            ]
            m.submodules += FFSynchronizer(self.scl_t.i, self.scl_i, init=1)
        else:
            # SCL output only
            m.d.comb += [
                self.scl_t.o  .eq(self.scl_o),
                self.scl_i    .eq(self.scl_o),
            ]

        # Additional signals for bus state detection
        scl_r = Signal(init=1)
        sda_r = Signal(init=1)
        m.d.sync += [
            scl_r.eq(self.scl_i),
            sda_r.eq(self.sda_i),
        ]
        m.d.comb += [
            self.sample .eq(~scl_r & self.scl_i),  # SCL rising edge
            self.setup  .eq(scl_r & ~self.scl_i),  # SCL falling edge
            self.start  .eq(self.scl_i & sda_r & ~self.sda_i),  # SDA fall, SCL high
            self.stop   .eq(self.scl_i & ~sda_r & self.sda_i),  # SDA rise, SCL high
        ]
        
        return m


class I2CInitiator(Elaboratable):
    """
    Simple I2C transaction initiator.

    Generates start and stop conditions, and transmits and receives octets.
    Clock stretching is supported.

    :param period_cyc:
        Bus clock period, as a multiple of system clock period.
    :type period_cyc: int
    :param clk_stretch:
        If true, SCL will be monitored for devices stretching the clock. Otherwise,
        only internally generated SCL is considered.
    :type clk_stretch: bool

    :attr busy:
        Busy flag. Low if the state machine is idle, high otherwise.
    :attr start:
        Start strobe. When ``busy`` is low, asserting ``start`` for one cycle generates
        a start or repeated start condition on the bus. Ignored when ``busy`` is high.
    :attr stop:
        Stop strobe. When ``busy`` is low, asserting ``stop`` for one cycle generates
        a stop condition on the bus. Ignored when ``busy`` is high.
    :attr write:
        Write strobe. When ``busy`` is low, asserting ``write`` for one cycle latches
        ``data_i`` and transmits it on the bus, after which the acknowledge bit
        from the bus is latched to ``ack_o``. Ignored when ``busy`` is high.
    :attr data_i:
        Data octet to be transmitted. Latched immediately after ``write`` is asserted.
    :attr ack_o:
        Received acknowledge bit.
    :attr read:
        Read strobe. 
        When ``busy`` is low, asserting ``read`` for one cycle receives an octet on 
        the bus and latches it to ``data_o``, after which the acknowledge bit is 
        asserted if ``ack_i`` is high. Ignored when ``busy`` is high.
    :attr data_o:
        Received data octet.
    :attr ack_i:
        Acknowledge bit to be transmitted. Latched immediately after ``read`` is asserted.
    """
    def __init__(self, pads, period_cyc, clk_stretch=True):
        self.bus         = I2CBusDriver(pads)
        self.period_cyc  = int(period_cyc)
        self.clk_stretch = clk_stretch

        self.busy   = Signal(init=1)
        self.start  = Signal()
        self.stop   = Signal()
        self.read   = Signal()
        self.data_i = Signal(8)
        self.ack_o  = Signal()
        self.write  = Signal()
        self.data_o = Signal(8)
        self.ack_i  = Signal()

    def elaborate(self, platform):
        m = Module()

        m.submodules.bus = bus = self.bus

        timer = Signal(range(self.period_cyc))
        stb   = Signal()

        with m.If((timer == 0) | ~self.busy):
            m.d.sync += timer.eq(self.period_cyc // 4)
        with m.Elif((not self.clk_stretch) | (bus.scl_o == bus.scl_i)):
            m.d.sync += timer.eq(timer - 1)

        m.d.comb += stb.eq(timer == 0)

        bitno   = Signal(range(8))
        r_shreg = Signal(8)
        w_shreg = Signal(8)
        r_ack   = Signal()
        
        with m.FSM() as fsm:
            self._fsm = fsm
            def scl_l(state, next_state, *exprs):
                with m.State(state):
                    with m.If(stb):
                        m.d.sync += self.bus.scl_o.eq(0)
                        m.next = next_state
                        m.d.sync += exprs

            def scl_h(state, next_state, *exprs):
                with m.State(state):
                    with m.If(stb):
                        m.d.sync += self.bus.scl_o.eq(1)
                    with m.Elif(self.bus.scl_o == 1):
                        with m.If((not self.clk_stretch) | (self.bus.scl_i == 1)):
                            m.next = next_state
                            m.d.sync += exprs

            def stb_x(state, next_state, *exprs, bit7_next_state=None):
                with m.State(state):
                    with m.If(stb):
                        m.next = next_state
                        if bit7_next_state is not None:
                            with m.If(bitno == 7):
                                m.next = bit7_next_state
                        m.d.sync += exprs

            with m.State("IDLE"):
                m.d.sync += self.busy.eq(1)
                with m.If(self.start):
                    with m.If(bus.scl_i & bus.sda_i):
                        m.next = "START-SDA-L"
                    with m.Elif(~bus.scl_i):
                        m.next = "START-SCL-H"
                    with m.Elif(bus.scl_i):
                        m.next = "START-SCL-L"
                with m.Elif(self.stop):
                    with m.If(bus.scl_i & ~bus.sda_o):
                        m.next = "STOP-SDA-H"
                    with m.Elif(~bus.scl_i):
                        m.next = "STOP-SCL-H"
                    with m.Elif(bus.scl_i):
                        m.next = "STOP-SCL-L"
                with m.Elif(self.write):
                    m.d.sync += w_shreg.eq(self.data_i)
                    m.next = "WRITE-DATA-SCL-L"
                with m.Elif(self.read):
                    m.d.sync += r_ack.eq(self.ack_i)
                    m.next = "READ-DATA-SCL-L"
                with m.Else():
                    m.d.sync += self.busy.eq(0)
            
            # start
            scl_l("START-SCL-L", "START-SDA-H")
            stb_x("START-SDA-H", "START-SCL-H",
                self.bus.sda_o.eq(1)
            )
            scl_h("START-SCL-H", "START-SDA-L")
            stb_x("START-SDA-L", "IDLE",
                self.bus.sda_o.eq(0)
            )
            # stop
            scl_l("STOP-SCL-L",  "STOP-SDA-L")
            stb_x("STOP-SDA-L",  "STOP-SCL-H",
                self.bus.sda_o.eq(0)
            )
            scl_h("STOP-SCL-H",  "STOP-SDA-H")
            stb_x("STOP-SDA-H",  "IDLE",
                self.bus.sda_o.eq(1)
            )
            # write data
            scl_l("WRITE-DATA-SCL-L", "WRITE-DATA-SDA-X")
            stb_x("WRITE-DATA-SDA-X", "WRITE-DATA-SCL-H",
                self.bus.sda_o.eq(w_shreg[7])
            )
            scl_h("WRITE-DATA-SCL-H", "WRITE-DATA-SDA-N",
                w_shreg.eq(Cat(C(0, 1), w_shreg[0:7]))
            )
            stb_x("WRITE-DATA-SDA-N", "WRITE-DATA-SCL-L",
                bitno.eq(bitno + 1),
                bit7_next_state="WRITE-ACK-SCL-L"
            )
            # write ack
            scl_l("WRITE-ACK-SCL-L", "WRITE-ACK-SDA-H")
            stb_x("WRITE-ACK-SDA-H", "WRITE-ACK-SCL-H",
                self.bus.sda_o.eq(1)
            )
            scl_h("WRITE-ACK-SCL-H", "WRITE-ACK-SDA-N",
                self.ack_o.eq(~self.bus.sda_i)
            )
            stb_x("WRITE-ACK-SDA-N", "IDLE")
            # read data
            scl_l("READ-DATA-SCL-L", "READ-DATA-SDA-H")
            stb_x("READ-DATA-SDA-H", "READ-DATA-SCL-H",
                self.bus.sda_o.eq(1)
            )
            scl_h("READ-DATA-SCL-H", "READ-DATA-SDA-N",
                r_shreg.eq(Cat(self.bus.sda_i, r_shreg[0:7]))
            )
            stb_x("READ-DATA-SDA-N", "READ-DATA-SCL-L",
                bitno.eq(bitno + 1),
                bit7_next_state="READ-ACK-SCL-L"
            )
            # read ack
            scl_l("READ-ACK-SCL-L", "READ-ACK-SDA-X")
            stb_x("READ-ACK-SDA-X", "READ-ACK-SCL-H",
                self.bus.sda_o.eq(~r_ack)
            )
            scl_h("READ-ACK-SCL-H", "READ-ACK-SDA-N",
                self.data_o.eq(r_shreg)
            )
            stb_x("READ-ACK-SDA-N", "IDLE")

        return m




#
# This file is part of LUNA.
#
""" Simple utility constructs for LUNA. """

from amaranth import Module, Signal, Cat

__all__ = [
    'rising_edge_detected', 'falling_edge_detected', 'any_edge_detected',
    'past_value_of'
]


def _single_edge_detector(m, signal, *, domain, edge='rising'):
    """ Generates and returns a signal that goes high for a cycle upon a given edge of a given signal. """

    # Create a one-cycle delayed version of our input signal.
    delayed = Signal()
    m.d[domain] += delayed.eq(signal)

    # And create a signal that detects edges on the relevant signal.
    edge_detected = Signal()
    if edge == 'falling':
        m.d.comb += edge_detected.eq(delayed & ~signal)
    elif edge == 'rising':
        m.d.comb += edge_detected.eq(~delayed & signal)
    elif edge == 'any':
        m.d.comb += edge_detected.eq(delayed != signal)
    else:
        raise ValueError("edge must be one of {rising,falling,any}")

    return edge_detected


def past_value_of(m, signal, *, domain):
    """ Generates and returns a signal that represents the value of another signal a cycle ago. """

    # Create a one-cycle delayed version of our input signal.
    delayed = Signal()
    m.d[domain] += delayed.eq(signal)

    return delayed


def rising_edge_detected(m, signal, *, domain="sync"):
    """ Generates and returns a signal that goes high for a cycle each rising edge of a given signal. """
    return _single_edge_detector(m, signal, edge='rising', domain=domain)


def falling_edge_detected(m, signal, *, domain="sync"):
    """ Generates and returns a signal that goes high for a cycle each rising edge of a given signal. """
    return _single_edge_detector(m, signal, edge='falling', domain=domain)


def any_edge_detected(m, signal, *, domain="sync"):
    """ Generates and returns a signal that goes high for a cycle each rising edge of a given signal. """
    return _single_edge_detector(m, signal, edge='any', domain=domain)




#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test            import usb_domain_test_case
from luna.gateware.test.usb2       import USBDeviceTest

from luna.gateware.usb.usb2        import USBPacketID
from luna.gateware.usb.usb2.device import USBDevice

from usb_protocol.emitters         import DeviceDescriptorCollection
from usb_protocol.types            import DescriptorTypes

class FullDeviceTest(USBDeviceTest):
    """ :meta private: """

    FRAGMENT_UNDER_TEST = USBDevice
    FRAGMENT_ARGUMENTS = {'handle_clocking': False}

    def traces_of_interest(self):
        return (
            self.utmi.tx_data,
            self.utmi.tx_valid,
            self.utmi.rx_data,
            self.utmi.rx_valid,
        )

    def initialize_signals(self):

        # Keep our device from resetting.
        yield self.utmi.line_state.eq(0b01)

        # Have our USB device connected.
        yield self.dut.connect.eq(1)

        # Pretend our PHY is always ready to accept data,
        # so we can move forward quickly.
        yield self.utmi.tx_ready.eq(1)


    def provision_dut(self, dut):
        self.descriptors = descriptors = DeviceDescriptorCollection()

        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = 0x1209
            d.idProduct          = 0x0001

            d.iManufacturer      = "LUNA"
            d.iProduct           = "Test Device"
            d.iSerialNumber      = "1234"

            d.bNumConfigurations = 1

        # Provide a core configuration descriptor for testing.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x01
                    e.wMaxPacketSize   = 512

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x81
                    e.wMaxPacketSize   = 512

        dut.add_standard_control_endpoint(descriptors)


    @usb_domain_test_case
    def test_enumeration(self):

        # Reference enumeration process (quirks merged from Linux, macOS, and Windows):
        # - Read 8 bytes of device descriptor.
        # - Read 64 bytes of device descriptor.
        # - Set address.
        # - Read exact device descriptor length.
        # - Read device qualifier descriptor, three times.
        # - Read config descriptor (without subordinates).
        # - Read language descriptor.
        # - Read Windows extended descriptors. [optional]
        # - Read string descriptors from device descriptor (wIndex=language id).
        # - Set configuration.
        # - Read back configuration number and validate.


        # Read 8 bytes of our device descriptor.
        handshake, data = yield from self.get_descriptor(DescriptorTypes.DEVICE, length=8)
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.DEVICE)[0:8])

        # Read 64 bytes of our device descriptor, no matter its length.
        handshake, data = yield from self.get_descriptor(DescriptorTypes.DEVICE, length=64)
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.DEVICE))

        # Send a nonsense request, and validate that it's stalled.
        handshake, data = yield from self.control_request_in(0x80, 30, length=10)
        self.assertEqual(handshake, USBPacketID.STALL)

        # Send a set-address request; we'll apply an arbitrary address 0x31.
        yield from self.set_address(0x31)
        self.assertEqual(self.address, 0x31)

        # Read our device descriptor.
        handshake, data = yield from self.get_descriptor(DescriptorTypes.DEVICE, length=18)
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.DEVICE))

        # Read our device qualifier descriptor.
        for _ in range(3):
            handshake, data = yield from self.get_descriptor(DescriptorTypes.DEVICE_QUALIFIER, length=10)
            self.assertEqual(handshake, USBPacketID.STALL)

        # Read our configuration descriptor (no subordinates).
        handshake, data = yield from self.get_descriptor(DescriptorTypes.CONFIGURATION, length=9)
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.CONFIGURATION)[0:9])

        # Read our configuration descriptor (with subordinates).
        handshake, data = yield from self.get_descriptor(DescriptorTypes.CONFIGURATION, length=32)
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.CONFIGURATION))

        # Read our string descriptors.
        for i in range(4):
            handshake, data = yield from self.get_descriptor(DescriptorTypes.STRING, index=i, length=255)
            self.assertEqual(handshake, USBPacketID.ACK)
            self.assertEqual(bytes(data), self.descriptors.get_descriptor_bytes(DescriptorTypes.STRING, index=i))

        # Set our configuration...
        status_pid = yield from self.set_configuration(1)
        self.assertEqual(status_pid, USBPacketID.DATA1)

        # ... and ensure it's applied.
        handshake, configuration = yield from self.get_configuration()
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(configuration, [1], "device did not accept configuration!")


class LongDescriptorTest(USBDeviceTest):
    """ :meta private: """

    FRAGMENT_UNDER_TEST = USBDevice
    FRAGMENT_ARGUMENTS = {'handle_clocking': False}

    def initialize_signals(self):

        # Keep our device from resetting.
        yield self.utmi.line_state.eq(0b01)

        # Have our USB device connected.
        yield self.dut.connect.eq(1)

        # Pretend our PHY is always ready to accept data,
        # so we can move forward quickly.
        yield self.utmi.tx_ready.eq(1)


    def provision_dut(self, dut):
        self.descriptors = descriptors = DeviceDescriptorCollection()

        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = 0x1209
            d.idProduct          = 0x0001

            d.iManufacturer      = "LUNA"
            d.iProduct           = "Test Device"
            d.iSerialNumber      = "1234"

            d.bNumConfigurations = 1

        # Provide a core configuration descriptor for testing.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0

                for n in range(15):

                    with i.EndpointDescriptor() as e:
                        e.bEndpointAddress = n
                        e.wMaxPacketSize   = 512

                    with i.EndpointDescriptor() as e:
                        e.bEndpointAddress = 0x80 | n
                        e.wMaxPacketSize   = 512

        dut.add_standard_control_endpoint(descriptors)

    @usb_domain_test_case
    def test_long_descriptor(self):
        descriptor = self.descriptors.get_descriptor_bytes(DescriptorTypes.CONFIGURATION)

        # Read our configuration descriptor (no subordinates).
        handshake, data = yield from self.get_descriptor(DescriptorTypes.CONFIGURATION, length=len(descriptor))
        self.assertEqual(handshake, USBPacketID.ACK)
        self.assertEqual(bytes(data), descriptor)
        self.assertEqual(len(data), len(descriptor))

    @usb_domain_test_case
    def test_descriptor_zlp(self):
        # Try requesting a long descriptor, but using a length that is a
        # multiple of the endpoint's maximum packet length. This should cause
        # the device to return some number of packets with the maximum packet
        # length, followed by a zero-length packet to terminate the
        # transaction.

        descriptor = self.descriptors.get_descriptor_bytes(DescriptorTypes.CONFIGURATION)

        # Try requesting a single and three max-sized packet.
        for factor in [1, 3]:
            request_length = self.max_packet_size_ep0 * factor
            handshake, data = yield from self.get_descriptor(DescriptorTypes.CONFIGURATION, length=request_length)
            self.assertEqual(handshake, USBPacketID.ACK)
            self.assertEqual(bytes(data), descriptor[0:request_length])
            self.assertEqual(len(data), request_length)



#
# This file is part of LUNA.
#
# Copyright (c) 2020-2025 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD--3-Clause
""" Endpoint interfaces for isochronous endpoints.

These interfaces provide interfaces for connecting streams or stream-like
interfaces to hosts via isochronous pipes.
"""

from amaranth      import *
from amaranth.lib  import stream

from ..endpoint    import EndpointInterface


class USBIsochronousStreamInEndpoint(Elaboratable):
    """ Isochronous endpoint that presents a stream-like interface.

    Used for repeatedly streaming data to a host from a stream-like interface.
    Intended to be useful as a transport for e.g. video or audio data.

    Attributes
    ----------
    stream: amaranth.lib.stream.Interface, input stream
        Full-featured stream interface that carries the data we'll transmit to the host.

    interface: EndpointInterface
        Communications link to our USB core.

    data_requested: Signal(), output
        Strobes, when a new packet starts

    frame_finished: Signal(), output
        Strobes immediately after the last byte in a frame has been transmitted

    bytes_in_frame: Signal(range(0, 3073)), input
        Specifies how many bytes will be transferred during this frame. If this is 0,
        a single ZLP will be emitted; for any other value one, two, or three packets
        will be generated, depending on the packet size. Latched in at the start of
        each frame.

        The maximum allowed value for this signal depends on the number of transfers
        per (micro)frame:
        - If this is a high-speed, high-throughput endpoint (descriptor indicates
          maxPacketSize > 512 and multiple transfers per microframe), then this value
          maxes out at (N * maxPacketSize), where N is the number of transfers per microframe.
        - For all other configurations, this must be <= the maximum packet size.

    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. Should match the wMaxPacketSize provided in the
        USB endpoint descriptor.
    """

    _MAX_FRAME_DATA = 1024 * 3

    def __init__(self, *, endpoint_number, max_packet_size):

        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O Port
        #
        self.interface      = EndpointInterface()
        self.stream         = stream.Interface(stream.Signature(unsigned(8)))
        self.data_requested = Signal()
        self.frame_finished = Signal()

        self.bytes_in_frame = Signal(range(0, self._MAX_FRAME_DATA + 1))

    def elaborate(self, platform):
        m = Module()

        # Shortcuts.
        interface        = self.interface
        tx_stream        = interface.tx
        new_frame        = interface.tokenizer.new_frame

        targeting_ep_num = (interface.tokenizer.endpoint == self._endpoint_number)
        targeting_us     = targeting_ep_num & interface.tokenizer.is_in
        data_requested   = targeting_us & interface.tokenizer.ready_for_response

        # Track our transmission state.
        bytes_left_in_frame  = Signal.like(self.bytes_in_frame)
        bytes_left_in_packet = Signal(range(0, self._max_packet_size + 1), init=self._max_packet_size - 1)
        next_data_pid        = Signal(2)
        tx_cnt               = Signal(range(0, self._MAX_FRAME_DATA))
        next_byte            = Signal.like(tx_cnt)

        m.d.comb += [
            tx_stream.payload        .eq(0),
            interface.tx_pid_toggle  .eq(next_data_pid)
        ]

        # Reset our state at the start of each frame.
        with m.If(new_frame):

            m.d.usb += [
                # Latch in how many bytes we'll be transmitting this frame.
                bytes_left_in_frame.eq(self.bytes_in_frame),

                # And start with a full packet to transmit.
                bytes_left_in_packet.eq(self._max_packet_size),
            ]

            # If it'll take more than two packets to send our data, start off with DATA2.
            # We'll follow with DATA1 and DATA0.
            with m.If(self.bytes_in_frame > (2 * self._max_packet_size)):
                m.d.usb += next_data_pid.eq(2)

            # Otherwise, if we need two, start with DATA1.
            with m.Elif(self.bytes_in_frame > self._max_packet_size):
                m.d.usb += next_data_pid.eq(1)

            # Otherwise, we'll start (and end) with DATA0.
            with m.Else():
                m.d.usb += next_data_pid.eq(0)

        #
        # Core sequencing FSM.
        #
        with m.FSM(domain="usb"):
            m.d.usb += self.frame_finished.eq(0)

            # IDLE -- the host hasn't yet requested data from our endpoint.
            with m.State("IDLE"):
                m.d.comb += next_byte.eq(0)
                m.d.usb  += [
                    tx_cnt.eq(0),
                    tx_stream.first.eq(0),
                ]

                # Once the host requests a packet from us...
                with m.If(data_requested):
                    # If we have data to send, send it.
                    with m.If(bytes_left_in_frame):
                        m.d.usb += tx_stream.first.eq(1)
                        m.next = "SEND_DATA"

                    # Otherwise, we'll send a ZLP.
                    with m.Else():
                        m.next = "SEND_ZLP"

                    # Strobe when a new packet starts.
                    m.d.comb += self.data_requested.eq(1)


            # SEND_DATA -- our primary data-transmission state; handles packet transmission
            with m.State("SEND_DATA"):
                last_byte_in_packet    = (bytes_left_in_packet <= 1)
                last_byte_in_frame     = (bytes_left_in_frame  <= 1)
                byte_terminates_send   = last_byte_in_packet | last_byte_in_frame

                m.d.comb += [
                    # Our data is always valid in this state...
                    tx_stream.valid .eq(1),
                    # ... and we're terminating if we're on the last byte of the packet or frame.
                    tx_stream.last  .eq(byte_terminates_send),
                ]

                # Strobe frame_finished one cycle after we're on the last byte of the frame.
                m.d.usb += self.frame_finished.eq(last_byte_in_frame)

                # Producer has data available.
                with m.If(self.stream.valid):
                    m.d.comb += tx_stream.payload.eq(self.stream.payload)

                # Don't advance ...
                m.d.comb += [
                    next_byte.eq(tx_cnt),
                    self.stream.ready.eq(0),
                ]
                m.d.usb += tx_cnt.eq(next_byte)

                # ... until our data is accepted.
                with m.If(tx_stream.ready):
                    m.d.usb += tx_stream.first.eq(0)

                    # Advance to the next byte in the frame ...
                    m.d.comb += [
                        self.stream.ready.eq(1),
                        next_byte.eq(tx_cnt + 1)
                    ]

                    # ... and mark the relevant byte as sent.
                    m.d.usb += [
                        bytes_left_in_frame   .eq(bytes_left_in_frame  - 1),
                        bytes_left_in_packet  .eq(bytes_left_in_packet - 1),
                    ]

                    # If we've just completed transmitting a packet, or we've
                    # just transmitted a full frame, end our transmission.
                    with m.If(byte_terminates_send):
                        m.d.usb += [
                            # Move to the next DATA pid, which is always one DATA PID less.
                            # [USB2.0: 5.9.2]. We'll reset this back to its maximum value when
                            # the next frame starts.
                            next_data_pid        .eq(next_data_pid - 1),

                            # Mark our next packet as being a full one.
                            bytes_left_in_packet .eq(self._max_packet_size),
                        ]
                        m.next = "IDLE"

            # SEND_ZLP -- sends a zero-length packet, and then return to idle.
            with m.State("SEND_ZLP"):
                # We'll request a ZLP by strobing LAST and VALID without strobing FIRST.
                m.d.comb += [
                    tx_stream.valid  .eq(1),
                    tx_stream.last   .eq(1),
                ]
                m.next = "IDLE"

        return m
